"""Module for Gens API."""

import logging
from pathlib import Path
from subprocess import CalledProcessError
from typing import Optional, Dict, List

from cg.constants.constants import FileFormat
from cg.exc import CaseNotFoundError
from cg.io.controller import ReadStream
from cg.utils import Process
from cg.utils.dict import get_list_from_dictionary

LOG = logging.getLogger(__name__)


class GensAPI:
    """API for Gens."""

    def __init__(self, config: Dict[str, Dict[str, str]]):
        self.binary_path: str = config["gens"]["binary_path"]
        self.config_path: str = config["gens"]["config_path"]
        self.process: Process = Process(
            binary=self.binary_path, config=self.config_path, config_parameter="--env-file"
        )
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set the dry run state."""
        self.dry_run = dry_run

    def load(
        self,
        baf_path: str,
        case_id: str,
        coverage_path: str,
        genome_build: str,
        sample_id: str,
    ) -> None:
        """Load Gens sample file paths into database."""
        load_params: Dict[str, str] = {
            "--sample-id": sample_id,
            "--genome-build": genome_build,
            "--baf": baf_path,
            "--coverage": coverage_path,
            "--case-id": case_id,
        }
        load_call_params: List[str] = ["load", "sample"] + get_list_from_dictionary(load_params)
        self.process.run_command(parameters=load_call_params, dry_run=self.dry_run)

    def __str__(self):
        return f"GensAPI(dry_run: {self.dry_run})"



from datetime import datetime, date
from pathlib import Path
from typing import List

import pytest
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.meta.upload.gisaid import GisaidAPI
from cg.meta.upload.gisaid.models import GisaidSample
from cg.models.cg_config import CGConfig, GisaidConfig, LimsConfig
from cg.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="config_object_no_gisaid_samples")
def fixture_config_object_no_gisaid_samples(
    config: dict, cg_config_object: CGConfig, base_store: Store
):
    cg_config_object.status_db_ = base_store
    cg_config_object.gisaid = GisaidConfig(**config["gisaid"])
    return cg_config_object


@pytest.fixture(name="config_object")
def fixture_config_object(
    cg_context: CGConfig, analysis_store: Store, real_housekeeper_api: HousekeeperAPI
):
    cg_context.status_db_ = analysis_store
    cg_context.housekeeper_api_ = real_housekeeper_api
    return cg_context


@pytest.fixture(scope="function")
def gisaid_api(config_object):
    """
    gisaid API fixture
    """

    return GisaidAPI(config=config_object)


@pytest.fixture(scope="function")
def gisaid_api_no_samples(config_object_no_gisaid_samples):
    """
    gisaid API fixture
    """
    return GisaidAPI(config=config_object_no_gisaid_samples)


@pytest.fixture(name="temp_result_file")
def fixture_temp_result_file() -> Path:
    file_path = Path("result_file_name")
    yield file_path
    try:
        file_path.unlink()
    except FileNotFoundError:
        pass


@pytest.fixture
def valid_gisiad_fasta_file() -> Path:
    """Get file path to valid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/valid_gisaid.fasta"
    return Path(file_path)


@pytest.fixture
def valid_housekeeper_fasta_file() -> Path:
    """Get file path to valid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/valid_housekeeper.fasta"
    return Path(file_path)


@pytest.fixture
def invalid_fasta_file() -> str:
    """Get file path to invalid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/invalid_housekeeper.fasta"
    return Path(file_path)


@pytest.fixture
def gisaid_case_id():
    return "gisaid_case"


@pytest.fixture
def four_gisaid_samples(gisaid_case_id: str, temp_result_file: Path) -> List[GisaidSample]:
    return [
        GisaidSample(
            case_id=gisaid_case_id,
            cg_lims_id=f"s{i}",
            covv_subm_sample_id=f"s{i}",
            submitter="test_submitter",
            fn=temp_result_file.name,
            covv_collection_date="2020-11-22",
            region="Stockholm",
            region_code="01",
            covv_orig_lab="Karolinska University Hospital",
            covv_orig_lab_addr="171 76 Stockholm, Sweden",
        )
        for i in range(1, 5)
    ]


def get_sample_attribute(lims_id: str, key: str):
    udfs = {
        "original_lab_address": "171 76 Stockholm, Sweden",
        "original_lab": "Karolinska University Hospital",
        "region_code": "01",
        "collection_date": "2020-11-22",
        "region": "Stockholm",
    }
    return udfs[key]


@pytest.fixture
def dummy_gisaid_sample(temp_result_file: Path) -> GisaidSample:
    return GisaidSample(
        case_id="dummy_id",
        cg_lims_id=f"dummy_sample",
        covv_subm_sample_id="test_name",
        submitter="test_submitter",
        fn=temp_result_file.name,
        covv_collection_date="2020-11-22",
        region="Stockholm",
        region_code="01",
        covv_orig_lab="Karolinska University Hospital",
        covv_orig_lab_addr="171 76 Stockholm, Sweden",
    )


@pytest.fixture
def bundle_with_four_samples(gisaid_case_id, timestamp):
    file = Path("tests/meta/upload/gisaid/fixtures/valid.fasta")
    files = [{"path": str(file.absolute()), "archive": False, "tags": ["consensus"]}]
    return {
        "name": gisaid_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture
def four_samples_csv() -> str:
    file_path = "tests/meta/upload/gisaid/fixtures/four_samples.csv"
    file = Path(file_path)
    return file.read_text()



"""Code for talking to Scout regarding uploads"""

import logging
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Optional

from cg.apps.scout.scout_export import ScoutExportCase, Variant
from cg.constants.constants import FileFormat
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.exc import ScoutUploadError
from cg.io.controller import ReadFile, ReadStream
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class ScoutAPI:

    """Interface to Scout."""

    def __init__(self, config):
        binary_path = config["scout"]["binary_path"]
        config_path = config["scout"]["config_path"]
        self.process = Process(binary=binary_path, config=config_path)

    def upload(self, scout_load_config: Path, threshold: int = 5, force: bool = False):
        """Load analysis of a new family into Scout."""

        scout_config: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=scout_load_config
        )
        scout_load_config_object: ScoutLoadConfig = ScoutLoadConfig(**scout_config)
        existing_case: Optional[ScoutExportCase] = self.get_case(
            case_id=scout_load_config_object.family
        )
        load_command = ["load", "case", str(scout_load_config)]
        if existing_case:
            if force or scout_load_config_object.analysis_date > existing_case.analysis_date:
                load_command.append("--update")
                LOG.info("update existing Scout case")
            else:
                existing_date = existing_case.analysis_date.date()
                LOG.warning("analysis of case already loaded: %s", existing_date)
                return
        LOG.debug("load new Scout case")
        self.process.run_command(load_command)
        LOG.debug("Case loaded successfully to Scout")

    def update_alignment_file(self, case_id: str, sample_id: str, alignment_path: Path):
        """Update alignment file for individual in case"""
        parameters = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind-id",
            sample_id,
            "--alignment-path",
            str(alignment_path),
        ]
        self.process.run_command(parameters=parameters)

    def export_panels(self, panels: List[str], build: str = GENOME_BUILD_37) -> List[str]:
        """Pass through to export of a list of gene panels.

        Return list of lines in bed format
        """
        export_panels_command = ["export", "panel", "--bed"]
        for panel_id in panels:
            export_panels_command.append(panel_id)

        if build:
            export_panels_command.extend(["--build", build])

        try:
            self.process.run_command(export_panels_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find panels")
            return []

        return [line for line in self.process.stdout_lines()]

    def get_genes(self, panel_id: str, build: str = None) -> list:
        """Fetch panel genes.

        Args:
            panel_id (str): unique id for the panel
            build (str): version of the panel. If 'None' latest version will be returned

        Returns:
            panel genes: panel genes list
        """
        # This can be run from CLI with `scout export panels <panel1> `
        export_panel_command = ["export", "panel", panel_id]
        if build:
            export_panel_command.extend(["--build", build])

        try:
            self.process.run_command(export_panel_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find panel %s", panel_id)
            return []

        panel_genes = []
        for gene_line in self.process.stdout_lines():
            if gene_line.startswith("#"):
                continue
            gene_info = gene_line.strip().split("\t")
            if not len(gene_info) > 1:
                continue
            panel_genes.append({"hgnc_id": int(gene_info[0]), "hgnc_symbol": gene_info[1]})

        return panel_genes

    def get_causative_variants(self, case_id: str) -> List[Variant]:
        """
        Get causative variants for a case
        """
        # These commands can be run with `scout export variants`
        get_causatives_command = ["export", "variants", "--json", "--case-id", case_id]
        try:
            self.process.run_command(get_causatives_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.warning("Could not find case %s in scout", case_id)
            return []
        variants: List[Variant] = []
        for variant_info in ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        ):
            variants.append(Variant(**variant_info))
        return variants

    def get_case(self, case_id: str) -> Optional[ScoutExportCase]:
        """Fetch a case from Scout"""
        cases: List[ScoutExportCase] = self.get_cases(case_id=case_id)
        if not cases:
            return None
        return cases[0]

    def get_cases(
        self,
        case_id: Optional[str] = None,
        reruns: bool = False,
        finished: bool = False,
        status: Optional[str] = None,
        days_ago: int = None,
    ) -> List[ScoutExportCase]:
        """Interact with cases existing in the database."""
        # These commands can be run with `scout export cases`
        get_cases_command = ["export", "cases", "--json"]
        if case_id:
            get_cases_command.extend(["--case-id", case_id])

        elif status:
            get_cases_command.extend(["--status", status])

        elif finished:
            get_cases_command.append("--finished")

        if reruns:
            LOG.info("Fetching cases that are reruns")
            get_cases_command.append("--reruns")

        if days_ago:
            get_cases_command.extend(["--within-days", str(days_ago)])

        try:
            self.process.run_command(get_cases_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find cases")
            return []

        cases: List[ScoutExportCase] = []
        for case_export in ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        ):
            LOG.info("Validating case %s", case_export.get("_id"))
            cases.append(ScoutExportCase(**case_export))
        return cases

    def get_solved_cases(self, days_ago: int) -> List[ScoutExportCase]:
        """
        Get cases solved within chosen timespan

        Args:
            days_ago (int): Maximum days ago a case has been solved

        Return:
            cases (list): list of cases
        """
        return self.get_cases(status="solved", days_ago=days_ago)

    def upload_delivery_report(self, report_path: str, case_id: str, update: bool = False) -> None:
        """Load a delivery report into a case in the database

        If the report already exists the function will exit.
        If the user want to load a report that is already in the database
        'update' has to be 'True'

        Args:
            report_path (string):       Path to delivery report
            case_id     (string):       Case identifier
            update      (bool):         If an existing report should be replaced

        Returns:
            updated_case(dict)

        """
        # This command can be run with `scout load delivery-report <CASE-ID> <REPORT-PATH>`
        upload_command = ["load", "delivery-report", case_id, report_path]

        if update:
            upload_command.append("--update")

        try:
            LOG.info("Uploading delivery report %s to case %s", report_path, case_id)
            self.process.run_command(upload_command)
        except CalledProcessError:
            LOG.warning("Something went wrong when uploading delivery report")

    def upload_fusion_report(
        self, case_id: str, report_path: str, research: bool, update: bool
    ) -> None:
        """Load a fusion report into a case in the database

        Args:
            report_path (string):       Path to delivery report
            case_id     (string):       Case identifier
            research    (bool):         Research report
            update      (bool):         If an existing report should be replaced
        Returns:
            Nothing
        """

        # This command can be run with
        # `scout load gene-fusion-report [-r] <case_id> <path/to/research_gene_fusion_report.pdf>`
        upload_command = ["load", "gene-fusion-report"]

        if research:
            upload_command.append("--research")

        if update:
            upload_command.append("--update")

        upload_command.extend([case_id, report_path])

        try:
            LOG.info("Uploading fusion report %s to case %s", report_path, case_id)
            self.process.run_command(upload_command)
        except CalledProcessError:
            raise ScoutUploadError("Something went wrong when uploading fusion report")

    def upload_splice_junctions_bed(self, file_path: str, case_id: str, customer_sample_id: str):
        """Load a splice junctions bed file into a case in the database.

        Args:
            file_path           (string):       Path to delivery report
            case_id             (string):       Case identifier
            customer_sample_id  (string):       Customers sample identifier
        Returns:
            updated_case(dict)

        """

        # This command can be run with
        # `scout update individual -c <case_id> -n <customer_sample_id> splice_junctions_bed
        #   <path/to/junction_file.bed>`
        upload_command = [
            "update",
            "individual",
            "-c",
            case_id,
            "-n",
            customer_sample_id,
            "splice_junctions_bed",
            file_path,
        ]

        try:
            LOG.info("Uploading splice junctions bed file %s to case %s", file_path, case_id)
            self.process.run_command(upload_command)
        except CalledProcessError:
            raise ScoutUploadError("Something went wrong when uploading splice junctions bed file")

    def upload_rna_coverage_bigwig(self, file_path: str, case_id: str, customer_sample_id: str):
        """Load a rna coverage bigwig file into a case in the database

        Args:
            file_path           (string):       Path to delivery report
            case_id             (string):       Case identifier
            customer_sample_id  (string):       Customers sample identifier
        Returns:
            updated_case(dict)

        """

        # This command can be run with
        # `scout update individual -c <case_id> -n <customer_sample_id> rna_coverage_bigwig
        #         <path/to/coverage_file.bigWig>`
        upload_command = [
            "update",
            "individual",
            "-c",
            case_id,
            "-n",
            customer_sample_id,
            "rna_coverage_bigwig",
            file_path,
        ]

        try:
            LOG.info("Uploading rna coverage bigwig file %s to case %s", file_path, case_id)
            self.process.run_command(upload_command)
        except CalledProcessError:
            raise ScoutUploadError("Something went wrong when uploading rna coverage bigwig file")



import datetime as dt
import logging

from cg.cli.clean import hk_case_bundle_files
from cg.constants.housekeeper_tags import WORKFLOW_PROTECTED_TAGS
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store import models
from cgmodels.cg.constants import Pipeline
from click.testing import CliRunner
from cg.utils.date import get_date_days_ago

from tests.store_helpers import StoreHelpers


def test_date_days_ago_zero_days():
    # GIVEN

    # WHEN calculating days ago 0 days ago
    result: dt.datetime = get_date_days_ago(0)

    # THEN result should be equal to today
    assert isinstance(result, dt.datetime)
    assert result.date() == dt.datetime.now().date()


def test_date_days_ago_one_days():
    # GIVEN we know what date it was yesterday
    days_ago = 1
    yesterday = dt.datetime.now() - dt.timedelta(days=days_ago)

    # WHEN calculating days ago 1 days ago
    result: dt.datetime = get_date_days_ago(days_ago)

    # THEN result should be equal to yesterday
    assert isinstance(result, dt.datetime)
    assert result.date() == yesterday.date()


def test_clean_hk_case_files_too_old(cli_runner: CliRunner, clean_context: CGConfig, caplog):
    # GIVEN no analysis in database
    days_ago = 365 * 100
    date_one_year_ago = get_date_days_ago(days_ago)
    context = clean_context
    assert not context.status_db.get_analyses_before_date(before=date_one_year_ago).all()

    # WHEN running the clean command
    caplog.set_level(logging.DEBUG)
    result = cli_runner.invoke(
        hk_case_bundle_files,
        ["--days-old", days_ago, "--dry-run"],
        obj=context,
        catch_exceptions=False,
    )

    # THEN it should be successful
    assert result.exit_code == 0
    # THEN it should report not having cleaned anything
    assert f"Process freed 0.0 GB" in caplog.text


def test_clean_hk_case_files_single_analysis(
    caplog,
    cg_context: CGConfig,
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    hk_bundle_data: dict,
    timestamp: dt.datetime,
):
    # GIVEN we have some analyses to clean
    context: CGConfig = cg_context
    store: Store = context.status_db
    days_ago: int = 1
    date_days_ago: dt.datetime = get_date_days_ago(days_ago)
    pipeline: Pipeline = Pipeline.MIP_DNA

    analysis: models.Analysis = helpers.add_analysis(
        store=store, started_at=date_days_ago, completed_at=date_days_ago, pipeline=pipeline
    )
    bundle_name: str = analysis.family.internal_id

    # GIVEN a housekeeper api with some files
    hk_bundle_data["name"] = bundle_name
    helpers.ensure_hk_bundle(cg_context.housekeeper_api, bundle_data=hk_bundle_data)

    # WHEN running the clean command
    caplog.set_level(logging.DEBUG)
    result = cli_runner.invoke(
        hk_case_bundle_files, ["--days-old", days_ago, "--dry-run"], obj=context
    )

    # THEN it should be successful
    assert result.exit_code == 0
    # THEN it should report some files to clean
    assert "Version found for" in caplog.text
    assert "has the tags" in caplog.text
    assert "has no protected tags" in caplog.text
    assert "found on disk" in caplog.text


def test_clean_hk_case_files_analysis_with_protected_tag(
    caplog,
    cg_context: CGConfig,
    cli_runner: CliRunner,
    helpers: StoreHelpers,
    hk_bundle_data: dict,
    timestamp: dt.datetime,
):
    # GIVEN we have some analyses to clean
    context: CGConfig = cg_context
    store: Store = context.status_db
    days_ago: int = 1
    date_days_ago: dt.datetime = get_date_days_ago(days_ago)
    pipeline: Pipeline = Pipeline.MIP_DNA

    analysis: models.Analysis = helpers.add_analysis(
        store=store, started_at=date_days_ago, completed_at=date_days_ago, pipeline=pipeline
    )
    bundle_name: str = analysis.family.internal_id

    # GIVEN a housekeeper api with some file with protected tags
    protected_tags = WORKFLOW_PROTECTED_TAGS[pipeline][0]
    hk_bundle_data["name"] = bundle_name
    hk_bundle_data["files"][0]["tags"] = protected_tags
    helpers.ensure_hk_bundle(cg_context.housekeeper_api, bundle_data=hk_bundle_data)

    # WHEN running the clean command
    caplog.set_level(logging.DEBUG)
    result = cli_runner.invoke(
        hk_case_bundle_files,
        ["--days-old", days_ago, "--dry-run"],
        obj=context,
        catch_exceptions=False,
    )

    # THEN it should be successful
    assert result.exit_code == 0
    # THEN it should report some files to clean
    assert "Version found for" in caplog.text
    assert "has the tags" in caplog.text
    assert "has the protected tag(s)" in caplog.text



"""Handler to find basic data objects"""
import datetime as dt
from typing import List, Optional, Callable

from sqlalchemy import desc
from sqlalchemy.orm import Query

from cg.store.models import (
    Application,
    ApplicationVersion,
    Bed,
    BedVersion,
    Customer,
    Collaboration,
    Organism,
    Panel,
    User,
)
from cg.store.api.base import BaseHandler
from cg.store.status_bed_filters import apply_bed_filter, BedFilters
from cg.store.status_bed_version_filters import BedVersionFilters, apply_bed_version_filter


class FindBasicDataHandler(BaseHandler):
    """Contains methods to find basic data model instances"""

    def application(self, tag: str) -> Application:
        """Fetch an application from the store."""
        return self.Application.query.filter_by(tag=tag).first()

    def applications(self, *, category=None, archived=None):
        """Fetch all applications."""
        records = self.Application.query
        if category:
            records = records.filter_by(prep_category=category)
        if archived is not None:
            records = records.filter_by(is_archived=archived)
        return records.order_by(self.Application.prep_category, self.Application.tag)

    def application_version(self, application: Application, version: int) -> ApplicationVersion:
        """Fetch an application version."""
        query = self.ApplicationVersion.query.filter_by(application=application, version=version)
        return query.first()

    def _get_bed_version_query(self) -> Query:
        """Return bed version query."""
        return self.BedVersion.query

    def get_bed_version_by_short_name(self, bed_version_short_name: str) -> BedVersion:
        """Return bed version with short name."""
        return apply_bed_version_filter(
            bed_versions=self._get_bed_version_query(),
            bed_version_short_name=bed_version_short_name,
            functions=[BedVersionFilters.get_bed_version_by_short_name],
        ).first()

    def _get_bed_query(self) -> Query:
        """Return bed query."""
        return self.Bed.query

    def get_beds(self) -> Query:
        """Returns all beds."""
        return self._get_bed_query()

    def get_bed_by_name(self, bed_name: str) -> Optional[Bed]:
        """Return bed by name."""
        return apply_bed_filter(
            beds=self._get_bed_query(), bed_name=bed_name, functions=[BedFilters.get_beds_by_name]
        ).first()

    def get_active_beds(self) -> Query:
        """Get all beds which are not archived."""
        bed_filter_functions: List[Callable] = [
            BedFilters.get_not_archived_beds,
            BedFilters.order_beds_by_name,
        ]
        return apply_bed_filter(beds=self._get_bed_query(), functions=bed_filter_functions)

    def get_latest_bed_version(self, bed_name: str) -> Optional[BedVersion]:
        """Return the latest bed version for a bed by supplied name."""
        bed: Optional[Bed] = self.get_bed_by_name(bed_name=bed_name)
        return bed.versions[-1] if bed and bed.versions else None

    def customer(self, internal_id: str) -> Customer:
        """Fetch a customer by internal id from the store."""
        return self.Customer.query.filter_by(internal_id=internal_id).first()

    def customers(self) -> List[Customer]:
        """Fetch all customers."""
        return self.Customer.query

    def collaboration(self, internal_id: str) -> Collaboration:
        """Fetch a customer group by internal id from the store."""
        return self.Collaboration.query.filter_by(internal_id=internal_id).first()

    def customer_by_id(self, id_: int) -> Customer:
        """Fetch a customer by id number from the store."""
        return self.Customer.query.filter_by(id=id_).first()

    def current_application_version(self, tag: str) -> Optional[ApplicationVersion]:
        """Fetch the current application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        if not application_obj:
            return None
        application_id = application_obj.id
        records = self.ApplicationVersion.query.filter_by(application_id=application_id)
        records = records.filter(self.ApplicationVersion.valid_from < dt.datetime.now())
        records = records.order_by(desc(self.ApplicationVersion.valid_from))

        return records.first()

    def latest_version(self, tag: str) -> Optional[ApplicationVersion]:
        """Fetch the latest application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        return (
            application_obj.versions[-1] if application_obj and application_obj.versions else None
        )

    def organism(self, internal_id: str) -> Organism:
        """Find an Organism by internal_id."""
        return self.Organism.query.filter_by(internal_id=internal_id).first()

    def organisms(self) -> List[Organism]:
        """Fetch all organisms"""
        return self.Organism.query.order_by(Organism.internal_id)

    def panel(self, abbrev):
        """Find a panel by abbreviation."""
        return self.Panel.query.filter_by(abbrev=abbrev).first()

    def panels(self):
        """Returns all panels."""
        return self.Panel.query.order_by(Panel.abbrev)

    def user(self, email: str) -> User:
        """Fetch a user from the store."""
        return self.User.query.filter_by(email=email).first()



from pathlib import Path

from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingNovaseqAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell


def test_rename_demux_result(
    demultiplexed_flow_cell_working_directory: Path,
    demultiplex_context: CGConfig,
    flow_cell: FlowCell,
):
    # GIVEN that this is the location of the demultiplex api
    demultiplex_context.demultiplex_api_.out_dir = demultiplexed_flow_cell_working_directory
    post_demux_api: DemuxPostProcessingNovaseqAPI = DemuxPostProcessingNovaseqAPI(
        config=demultiplex_context
    )
    demux_dir: Path = demultiplexed_flow_cell_working_directory
    demux_results: DemuxResults = DemuxResults(
        demux_dir=demux_dir, flow_cell=flow_cell, bcl_converter="bcl2fastq"
    )
    # GIVEN that there are no projects with the correct file name
    assert not list(demux_results.projects)

    # GIVEN the location of a demultiplexed flowcell that is not finished
    assert list(demux_results.raw_projects)

    # WHEN renaming the files
    post_demux_api.rename_files(demux_results=demux_results)

    # THEN assert that the files have been renamed
    assert list(demux_results.projects)
    assert not list(demux_results.raw_projects)



from datetime import datetime

from alchy import Query
from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.store import Store, models
from cg.store.status_analysis_filters import (
    get_valid_analyses_in_production,
    get_analyses_with_pipeline,
    get_completed_analyses,
    get_not_completed_analyses,
    get_filter_uploaded_analyses,
    get_not_uploaded_analyses,
    get_analyses_with_delivery_report,
    get_analyses_without_delivery_report,
    get_report_analyses_by_pipeline,
    order_analyses_by_uploaded_at,
    order_analyses_by_completed_at,
)
from tests.store_helpers import StoreHelpers


def test_get_valid_analyses_in_production(
    base_store: Store,
    helpers: StoreHelpers,
    case_obj: models.Family,
    timestamp_now: datetime,
    old_timestamp: datetime,
):
    """Test that an expected analysis is returned when it has a production valid completed_at date."""

    # GIVEN a set of mock analyses
    analysis: models.Analysis = helpers.add_analysis(store=base_store, completed_at=timestamp_now)
    outdated_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, case=case_obj, completed_at=old_timestamp
    )
    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN retrieving valid in production analyses
    analyses: Query = get_valid_analyses_in_production(analyses_query)

    # THEN only the up-to-date analysis should be returned
    assert analysis in analyses
    assert outdated_analysis not in analyses


def test_get_analyses_with_pipeline(
    base_store: Store, helpers: StoreHelpers, case_obj: models.Family
):
    """Test analyses filtering by pipeline."""

    # GIVEN a set of mock analyses
    balsamic_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, pipeline=Pipeline.BALSAMIC
    )
    mip_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, case=case_obj, pipeline=Pipeline.MIP_DNA
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN extracting the analyses
    analyses: Query = get_analyses_with_pipeline(analyses_query, pipeline=Pipeline.BALSAMIC)

    # THEN only the BALSAMIC analysis should be retrieved
    assert balsamic_analysis in analyses
    assert mip_analysis not in analyses


def test_get_completed_analyses(base_store: Store, helpers: StoreHelpers, timestamp_now: datetime):
    """Test filtering of completed analyses."""

    # GIVEN a mock analysis
    analysis: models.Analysis = helpers.add_analysis(store=base_store, completed_at=timestamp_now)

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN retrieving the completed analyses
    analyses: Query = get_completed_analyses(analyses_query)

    # THEN the completed analysis should be obtained
    assert analysis in analyses


def test_get_not_completed_analyses(base_store: Store, helpers: StoreHelpers):
    """Test filtering of ongoing analyses."""

    # GIVEN a mock not completed analysis
    analysis_not_completed: models.Analysis = helpers.add_analysis(
        store=base_store, completed_at=None
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN retrieving the not completed analyses
    analyses: Query = get_not_completed_analyses(analyses_query)

    # THEN the expected analysis should be retrieved
    assert analysis_not_completed in analyses


def test_get_filter_uploaded_analyses(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analysis with an uploaded_at field."""

    # GIVEN a mock uploaded analysis
    analysis: models.Analysis = helpers.add_analysis(store=base_store, uploaded_at=timestamp_now)

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN calling the upload filtering function
    analyses: Query = get_filter_uploaded_analyses(analyses_query)

    # THEN the uploaded analysis should be retrieved
    assert analysis in analyses


def test_get_not_uploaded_analyses(base_store: Store, helpers: StoreHelpers):
    """Test filtering of analysis that has not been uploaded."""

    # GIVEN a mock not uploaded analysis
    not_uploaded_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, uploaded_at=None
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN calling the upload filtering function
    analyses: Query = get_not_uploaded_analyses(analyses_query)

    # THEN the uploaded analysis should be retrieved
    assert not_uploaded_analysis in analyses


def test_get_analyses_with_delivery_report(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analysis with a delivery report generated."""

    # GIVEN an analysis with a delivery report
    analysis: models.Analysis = helpers.add_analysis(
        store=base_store, delivery_reported_at=timestamp_now
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN calling the delivery report analysis filtering function
    analyses: Query = get_analyses_with_delivery_report(analyses_query)

    # THEN the analysis containing the delivery report should be extracted
    assert analysis in analyses


def test_get_analyses_without_delivery_report(base_store: Store, helpers: StoreHelpers):
    """Test filtering of analysis without a delivery report generated."""

    # GIVEN an analysis with a delivery report
    analysis_without_delivery_report: models.Analysis = helpers.add_analysis(
        store=base_store, delivery_reported_at=None
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN calling the delivery report analysis filtering function
    analyses: Query = get_analyses_without_delivery_report(analyses_query)

    # THEN the analysis without a delivery report should be extracted
    assert analysis_without_delivery_report in analyses


def test_get_report_analyses_by_pipeline(
    base_store: Store, helpers: StoreHelpers, case_obj: models.Family
):
    """Test filtering delivery report related analysis by pipeline."""

    # GIVEN a set of mock analysis
    balsamic_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, pipeline=Pipeline.BALSAMIC
    )
    fluffy_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, case=case_obj, pipeline=Pipeline.FLUFFY
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN filtering delivery report related analyses
    analyses: Query = get_report_analyses_by_pipeline(analyses_query)

    # THEN only the delivery report supported analysis should be retrieved
    assert balsamic_analysis in analyses
    assert fluffy_analysis not in analyses


def test_order_analyses_by_completed_at(
    base_store: Store,
    helpers: StoreHelpers,
    case_obj: models.Family,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
):
    """Test sorting of analyses by the completed_at field."""

    # GIVEN a set of mock analyses
    new_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, completed_at=timestamp_now
    )
    old_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, case=case_obj, completed_at=timestamp_yesterday
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN ordering the analyses by the completed_at field
    analyses: Query = order_analyses_by_completed_at(analyses_query)

    # THEN the oldest analysis should be the first one in the list
    assert old_analysis == analyses.all()[0]
    assert new_analysis == analyses.all()[1]


def test_order_analyses_by_uploaded_at(
    base_store: Store,
    helpers: StoreHelpers,
    case_obj: models.Family,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
):
    """Test sorting of analyses by the uploaded_at field."""

    # GIVEN a set of mock analyses
    new_analysis: models.Analysis = helpers.add_analysis(
        store=base_store, completed_at=timestamp_now, uploaded_at=timestamp_now
    )
    old_analysis: models.Analysis = helpers.add_analysis(
        store=base_store,
        case=case_obj,
        completed_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
    )

    # GIVEN an analysis query
    analyses_query: Query = base_store.latest_analyses()

    # WHEN ordering the analyses by the uploaded_at field
    analyses: Query = order_analyses_by_uploaded_at(analyses_query)

    # THEN the oldest analysis should be the first one in the list
    assert old_analysis == analyses.all()[0]
    assert new_analysis == analyses.all()[1]



"""Tests for the models in scout load config"""
from pathlib import Path
from typing import Any

import pytest
from pydantic import ValidationError

from cg.models.scout import scout_load_config
from cg.models.scout.scout_load_config import MipLoadConfig, ScoutMipIndividual
from tests.apps.scout.conftest import SCOUT_INDIVIDUAL


@pytest.mark.parametrize("key, value", list(SCOUT_INDIVIDUAL.items()))
def test_validate_scout_individual_attributes(scout_individual: dict, key: str, value: Any):
    """Test to validate that all attributes of a ScoutMipIndividual are correctly set."""
    # GIVEN some sample information
    # WHEN instantiating a ScoutMipIndividual
    ind_obj: ScoutMipIndividual = scout_load_config.ScoutMipIndividual(**scout_individual)

    # THEN assert that the attribute is set correctly
    assert getattr(ind_obj, key) == value


def test_instantiate_empty_mip_config():
    """Tests whether a MipLoadConfig can be instantiate without arguments."""
    # GIVEN nothing

    # WHEN instantiating a empty mip load config
    config: MipLoadConfig = scout_load_config.MipLoadConfig()

    # THEN assert it is possible to instantiate without any information
    assert isinstance(config, scout_load_config.ScoutLoadConfig)


def test_set_mandatory_to_none():
    """The scout load config object should validate fields as they are set.

    This test will check that a value error is raised when a mandatory field is set to None.
    """
    # GIVEN a load config object
    config: MipLoadConfig = scout_load_config.MipLoadConfig()

    # WHEN setting a mandatory field to None
    with pytest.raises(ValidationError):
        # THEN assert a validation error was raised
        config.vcf_snv = None



"""MIP-RNA upload API."""

import datetime as dt
import logging

import click

from cg.cli.upload.clinical_delivery import clinical_delivery
from cg.cli.upload.scout import upload_rna_to_scout
from cg.constants import DataDelivery
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Family

LOG = logging.getLogger(__name__)


class MipRNAUploadAPI(UploadAPI):
    """MIP-RNA upload API."""

    def __init__(self, config: CGConfig):
        self.analysis_api: MipRNAAnalysisAPI = MipRNAAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Family, restart: bool) -> None:
        """Uploads MIP-RNA analysis data and files."""

        analysis: Analysis = case.analyses[0]
        self.update_upload_started_at(analysis=analysis)

        # Clinical delivery upload
        ctx.invoke(clinical_delivery, case_id=case.internal_id)

        # Scout specific upload
        if DataDelivery.SCOUT in case.data_delivery:
            result: int = ctx.invoke(upload_rna_to_scout, case_id=case.internal_id)
            if result == 0:
                LOG.info(
                    f"Upload of case {case.internal_id} was successful. Setting uploaded at to {dt.datetime.now()}"
                )
                self.update_uploaded_at(analysis)
            else:
                raise RuntimeError(f"Upload to Scout failed for sample {case.internal_id}")
        else:
            LOG.warning(
                f"There is nothing to upload to Scout for case {case.internal_id} and "
                f"the specified data delivery ({case.data_delivery})"
            )



"""Test methods for cg/cli/set/families"""
import logging

import pytest
from cg.cli.set.families import families
from cg.models.cg_config import CGConfig
from cg.store import Store, models

SUCCESS = 0


@pytest.mark.parametrize("identifier_key", ["original_ticket", "order"])
def test_set_families_by_sample_identifiers(
    cli_runner, base_context: CGConfig, identifier_key, helpers, caplog, ticket
):
    # GIVEN a database with a case with a sample
    base_store: Store = base_context.status_db
    sample_obj: models.Sample = helpers.add_sample(base_store)
    sample_obj.original_ticket = ticket
    sample_obj.order = "An order"
    case: models.Family = helpers.add_case(base_store)
    helpers.add_relationship(base_store, sample=sample_obj, case=case)
    identifier_value = getattr(sample_obj, identifier_key)

    caplog.set_level(logging.INFO)

    # WHEN calling set families with valid sample identifiers
    cli_runner.invoke(
        families,
        ["--sample-identifier", identifier_key, identifier_value],
        obj=base_context,
    )

    # THEN it should name the case to be changed
    assert case.internal_id in caplog.text
    assert case.name in caplog.text



import logging
import re
from typing import Iterable, List, Optional

import click
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from tabulate import tabulate

LOG = logging.getLogger(__name__)
ANALYSIS_HEADERS = ["Analysis Date", "Pipeline", "Version"]
FAMILY_HEADERS = ["Family", "Name", "Customer", "Priority", "Panels", "Action"]
FLOWCELL_HEADERS = ["Flowcell", "Type", "Sequencer", "Date", "Archived?", "Status"]
LINK_HEADERS = ["Sample", "Mother", "Father"]
SAMPLE_HEADERS = ["Sample", "Name", "Customer", "Application", "State", "Priority", "External?"]


@click.group(invoke_without_command=True)
@click.option("-i", "--identifier", help="made a guess what type you are looking for")
@click.pass_context
def get(context: click.Context, identifier: Optional[str]):
    """Get information about records in the database."""
    if identifier:
        if re.match(r"^[A-Z]{3}[0-9]{4,5}[A-Z]{1}[1-9]{1,3}$", identifier):
            context.invoke(sample, sample_ids=[identifier])
        elif re.match(r"^[a-z]*$", identifier):
            # try family information
            context.invoke(family, family_ids=[identifier])
        elif re.match(r"^[HC][A-Z0-9]{8}$", identifier):
            # try flowcell information
            context.invoke(flowcell, flowcell_id=identifier)
        else:
            LOG.error(f"{identifier}: can't predict identifier")
            raise click.Abort


@get.command()
@click.option("--families/--no-families", default=True, help="display related families")
@click.option("-hf", "--hide-flowcell", is_flag=True, help="hide related flowcells")
@click.argument("sample_ids", nargs=-1)
@click.pass_context
def sample(context: click.Context, families: bool, hide_flowcell: bool, sample_ids: List[str]):
    """Get information about a sample."""
    status_db: Store = context.obj.status_db
    for sample_id in sample_ids:
        LOG.debug("%s: get info about sample", sample_id)
        sample_obj: models.Sample = status_db.sample(sample_id)
        if sample_obj is None:
            LOG.warning(f"{sample_id}: sample doesn't exist")
            continue
        row = [
            sample_obj.internal_id,
            sample_obj.name,
            sample_obj.customer.internal_id,
            sample_obj.application_version.application.tag,
            sample_obj.state,
            sample_obj.priority_human,
            "Yes" if sample_obj.application_version.application.is_external else "No",
        ]
        click.echo(tabulate([row], headers=SAMPLE_HEADERS, tablefmt="psql"))
        if families:
            family_ids: List[str] = [link_obj.family.internal_id for link_obj in sample_obj.links]
            context.invoke(family, family_ids=family_ids, samples=False)
        if not hide_flowcell:
            for flowcell_obj in sample_obj.flowcells:
                LOG.debug(f"{flowcell_obj.name}: get info about flowcell")
                context.invoke(flowcell, flowcell_id=flowcell_obj.name, samples=False)


@get.command()
@click.argument("case_id")
@click.pass_obj
def analysis(context: CGConfig, case_id: str):
    """Get information about case analysis."""
    status_db: Store = context.status_db
    case_obj: models.Family = status_db.family(case_id)
    if case_obj is None:
        LOG.error("%s: case doesn't exist", case_id)
        raise click.Abort

    LOG.debug("%s: get info about case analysis", case_obj.internal_id)

    for analysis_obj in case_obj.analyses:
        row = [
            analysis_obj.started_at,
            analysis_obj.pipeline,
            analysis_obj.pipeline_version,
        ]
        click.echo(tabulate([row], headers=ANALYSIS_HEADERS, tablefmt="psql"))


@get.command()
@click.argument("family_id")
@click.pass_obj
def relations(context: CGConfig, family_id: str):
    """Get information about family relations."""
    status_db: Store = context.status_db
    case_obj: models.Family = status_db.family(family_id)
    if case_obj is None:
        LOG.error("%s: family doesn't exist", family_id)
        raise click.Abort

    LOG.debug("%s: get info about family relations", case_obj.internal_id)

    for link_obj in case_obj.links:
        row = [
            link_obj.sample.internal_id if link_obj.sample else "",
            link_obj.mother.internal_id if link_obj.mother else "",
            link_obj.father.internal_id if link_obj.father else "",
        ]
        click.echo(tabulate([row], headers=LINK_HEADERS, tablefmt="psql"))


@get.command()
@click.option("-c", "--customer", help="internal id for customer to filter by")
@click.option("-n", "--name", is_flag=True, help="search family by name")
@click.option("--samples/--no-samples", default=True, help="display related samples")
@click.option("--relate/--no-relate", default=True, help="display relations to samples")
@click.option("--analyses/--no-analyses", default=True, help="display related analyses")
@click.argument("family_ids", nargs=-1)
@click.pass_context
def family(
    context: click.Context,
    customer: str,
    name: bool,
    samples: bool,
    relate: bool,
    analyses: bool,
    family_ids: List[str],
):
    """Get information about a family."""
    status_db: Store = context.obj.status_db
    cases: List[models.Family] = []
    if name:
        customer_obj: models.Customer = status_db.customer(customer)
        if customer_obj is None:
            LOG.error(f"{customer}: customer not found")
            raise click.Abort
        cases: Iterable[models.Family] = status_db.families(
            customers=[customer_obj], enquiry=family_ids[-1]
        )
    else:
        for family_id in family_ids:
            case_obj: models.Family = status_db.family(family_id)
            if case_obj is None:
                LOG.error(f"{family_id}: family doesn't exist")
                raise click.Abort
            cases.append(case_obj)

    for case_obj in cases:
        LOG.debug(f"{case_obj.internal_id}: get info about family")
        row: List[str] = [
            case_obj.internal_id,
            case_obj.name,
            case_obj.customer.internal_id,
            case_obj.priority_human,
            ", ".join(case_obj.panels),
            case_obj.action or "NA",
        ]
        click.echo(tabulate([row], headers=FAMILY_HEADERS, tablefmt="psql"))
        if relate:
            context.invoke(relations, family_id=case_obj.internal_id)
        if samples:
            sample_ids: List[str] = [link_obj.sample.internal_id for link_obj in case_obj.links]
            context.invoke(sample, sample_ids=sample_ids, families=False)
        if analyses:
            context.invoke(analysis, case_id=case_obj.internal_id)


@get.command()
@click.option("--samples/--no-samples", default=True, help="display related samples")
@click.argument("flowcell_id")
@click.pass_context
def flowcell(context: click.Context, samples: bool, flowcell_id: str):
    """Get information about a flowcell and the samples on it."""
    status_db: Store = context.obj.status_db
    flowcell_obj: models.Flowcell = status_db.get_flow_cell(flowcell_id)
    if flowcell_obj is None:
        LOG.error(f"{flowcell_id}: flowcell not found")
        raise click.Abort
    row: List[str] = [
        flowcell_obj.name,
        flowcell_obj.sequencer_type,
        flowcell_obj.sequencer_name,
        flowcell_obj.sequenced_at.date(),
        flowcell_obj.archived_at.date() if flowcell_obj.archived_at else "No",
        flowcell_obj.status,
    ]
    click.echo(tabulate([row], headers=FLOWCELL_HEADERS, tablefmt="psql"))
    if samples:
        sample_ids: List[str] = [sample_obj.internal_id for sample_obj in flowcell_obj.samples]
        if sample_ids:
            context.invoke(sample, sample_ids=sample_ids, families=False)
        else:
            LOG.warning("no samples found on flowcell")



""" Tests for cg.apps.mutacc_auto module"""

import subprocess

import pytest
from cg.apps.mutacc_auto import MutaccAutoAPI, run_command


def test_instatiate(mutacc_config):
    """Test to instatiate an MutaccAutoAPI object"""
    # GIVEN a dict with config data to

    # WHEN instatiating a mutacc_auto API

    mutacc_api = MutaccAutoAPI(config=mutacc_config)

    # THEN the object attributes are set correctly

    assert mutacc_api.mutacc_auto_config == mutacc_config["mutacc_auto"]["config_path"]
    assert mutacc_api.mutacc_auto_binary == mutacc_config["mutacc_auto"]["binary_path"]
    assert mutacc_api.mutacc_padding == mutacc_config["mutacc_auto"]["padding"]
    assert mutacc_api.base_call == [
        mutacc_config["mutacc_auto"]["binary_path"],
        "--config-file",
        mutacc_config["mutacc_auto"]["config_path"],
    ]


def test_extract_reads(mutacc_auto_api, mocker):
    """
    Test extract_reads method
    """
    # GIVEN a mutacc-auto api a case, and list of variants
    case = {}
    variants = []
    # WHEN calling MutaccAutoAPI.extract_reads

    # patch the run_command funciton
    mock_run_command = mocker.patch("cg.apps.mutacc_auto.run_command")
    mutacc_auto_api.extract_reads(case, variants)

    # THEN the command and its options should be as expected
    expected_call = mutacc_auto_api.base_call + [
        "extract",
        "--variants",
        str(variants),
        "--case",
        str(case),
        "--padding",
        str(mutacc_auto_api.mutacc_padding),
    ]

    mock_run_command.assert_called_with(expected_call)


def test_import_reads(mutacc_auto_api, mocker):
    """
    Test import_reads method
    """
    # GIVEN a mutacc-auto api

    # WHEN calling MutaccAutoAPI.import_reads

    mock_run_command = mocker.patch("cg.apps.mutacc_auto.run_command")
    mutacc_auto_api.import_reads()

    # THEN command and its options should be as expected
    expected_call = mutacc_auto_api.base_call + ["import"]
    mock_run_command.assert_called_with(expected_call)


def test_run_command(mock_failed_process, mocker):
    """
    Test run_command function
    """
    # WITH a mocked subprocess.run function
    mocker.patch.object(subprocess, "run")
    subprocess.run.return_value = mock_failed_process

    # WHEN trying to run a command with run_command

    # Then Error subprocess.CalledProcessError is raised
    with pytest.raises(subprocess.CalledProcessError):
        run_command(["command"])



"""Loqusdb input files models."""

import logging
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, validator


LOG = logging.getLogger(__name__)


def validate_observations_file(file: Path) -> Path:
    """Check if file exists in file system."""
    if file and not file.exists():
        LOG.error(f"File {file} could not be found")
        raise FileNotFoundError
    return file


class ObservationsInputFiles(BaseModel):
    """Model for validating Loqusdb input files."""

    snv_vcf_path: Path
    sv_vcf_path: Optional[Path] = None

    _ = validator("snv_vcf_path", "sv_vcf_path", always=False, allow_reuse=True)(
        validate_observations_file
    )


class MipDNAObservationsInputFiles(ObservationsInputFiles):
    """Model for validating rare disease Loqusdb input files."""

    profile_vcf_path: Path
    family_ped_path: Path

    _ = validator("family_ped_path", "profile_vcf_path", always=False, allow_reuse=True)(
        validate_observations_file
    )


class BalsamicObservationsInputFiles(ObservationsInputFiles):
    """Model for validating cancer Loqusdb input files."""

    snv_all_vcf_path: Path

    _ = validator("snv_all_vcf_path", always=False, allow_reuse=True)(validate_observations_file)



from datetime import datetime, timedelta

from cg.utils.date import get_timedelta_from_date, get_date_days_ago


def test_get_date_days_ago(timestamp_now: datetime):
    """Test returning a date corresponding to number of days ago."""

    # GIVEN days ago

    # WHEN calling the function
    date: datetime = get_date_days_ago(days_ago=1)

    # THEN assert the return should be a date
    assert isinstance(date, datetime)

    # Then the date returned should be less than then today
    assert date < timestamp_now


def test_get_timedelta_from_date(timestamp_yesterday: datetime):
    """Test returning the number of days ago corresponding to a given date."""

    # GIVEN a date

    # WHEN calling the function
    age: timedelta = get_timedelta_from_date(date=timestamp_yesterday)

    # THEN assert the return should be a time delta
    assert isinstance(age, timedelta)

    # Then the age in days should be yesterday
    assert age.days == 1



"""Fixtures for the upload Scout API tests."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Generator, List

import pytest
from housekeeper.store import models as hk_models
from housekeeper.store.models import Version

from cg.constants import DataDelivery, FileExtensions, Pipeline
from cg.constants.constants import FileFormat, PrepCategory
from cg.constants.sequencing import SequencingMethod
from cg.io.controller import ReadFile
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import MipLoadConfig
from cg.store import Store, models
from cg.store.models import Analysis, Family

# Mocks
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.scout import MockScoutAPI
from tests.store_helpers import StoreHelpers

LOG = logging.getLogger(__name__)
SNV_FILE_NAME: str = f"snv{FileExtensions.VCF}"


@pytest.fixture(name="rna_case_id")
def fixture_rna_case_id() -> str:
    """Return a rna case id."""
    return "affirmativecat"


@pytest.fixture(name="dna_case_id")
def fixture_dna_case_id(case_id: str) -> str:
    """Return a DNA case id."""
    return case_id


@pytest.fixture(name="rna_sample_son_id")
def fixture_rna_sample_son_id() -> str:
    """Return a son RNA sample id."""
    return "rna_son"


@pytest.fixture(name="rna_sample_daughter_id")
def fixture_rna_sample_daughter_id() -> str:
    """Return a daughter RNA sample id."""
    return "rna_daughter"


@pytest.fixture(name="rna_sample_mother_id")
def fixture_rna_sample_mother_id() -> str:
    """Return a mother RNA sample id."""
    return "rna_mother"


@pytest.fixture(name="rna_sample_father_id")
def fixture_rna_sample_father_id() -> str:
    """Return a father RNA sample id."""
    return "rna_father"


@pytest.fixture(name="dna_sample_son_id")
def fixture_dna_sample_son_id() -> str:
    """Return a son DNA sample id."""
    return "dna_son"


@pytest.fixture(name="dna_sample_daughter_id")
def fixture_dna_sample_daughter_id() -> str:
    """Return a daughter DNA sample id."""
    return "dna_daughter"


@pytest.fixture(name="dna_sample_mother_id")
def fixture_dna_sample_mother_id() -> str:
    """Return a mother DNA sample id."""
    return "dna_mother"


@pytest.fixture(name="dna_sample_father_id")
def fixture_dna_sample_father_id() -> str:
    """Return a father DNA sample id."""
    return "dna_father"


@pytest.fixture(name="another_sample_id")
def fixture_another_sample_id() -> str:
    """Return another sample id."""
    return "another_sample_id"


@pytest.fixture(name="another_rna_sample_id")
def fixture_another_rna_sample_id() -> str:
    """Return another RNA sample id."""
    return "another_rna_sample_id"


@pytest.fixture(name="rna_store")
def fixture_rna_store(
    base_store: Store,
    helpers: StoreHelpers,
    rna_case_id: str,
    dna_case_id: str,
) -> Store:
    """Populate store with a RNA case that is connected to a DNA case via sample.subject_id."""

    store: Store = base_store

    # an existing RNA case with related sample
    rna_case = helpers.ensure_case(
        store=store,
        name="rna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_RNA,
        data_delivery=DataDelivery.SCOUT,
    )
    rna_case.internal_id = rna_case_id

    rna_sample_son = helpers.add_sample(
        store=store,
        name="rna_son",
        subject_id="son",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_daughter = helpers.add_sample(
        store=store,
        name="rna_daughter",
        subject_id="daughter",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_mother = helpers.add_sample(
        store=store,
        name="rna_mother",
        subject_id="mother",
        application_type=SequencingMethod.WTS,
    )
    rna_sample_father = helpers.add_sample(
        store=store,
        name="rna_father",
        subject_id="father",
        application_type=SequencingMethod.WTS,
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_son,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_daughter,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_mother, case=rna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_father, case=rna_case, status="affected"
    )

    for link in rna_case.links:
        link.sample.internal_id = link.sample.name

    # an existing DNA case with related sample
    dna_case = helpers.ensure_case(
        store=store,
        name="dna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_DNA,
        data_delivery=DataDelivery.SCOUT,
    )
    dna_case.internal_id = dna_case_id

    dna_sample_son = helpers.add_sample(
        store=store,
        name="dna_son",
        subject_id="son",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_daughter = helpers.add_sample(
        store=store,
        name="dna_daughter",
        subject_id="daughter",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_mother = helpers.add_sample(
        store=store,
        name="dna_mother",
        subject_id="mother",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    dna_sample_father = helpers.add_sample(
        store=store,
        name="dna_father",
        subject_id="father",
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_son,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_daughter,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_mother, case=dna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_father, case=dna_case, status="affected"
    )

    for link in dna_case.links:
        link.sample.internal_id = link.sample.name

    store.commit()
    return store


@pytest.fixture(name="lims_family")
def fixture_lims_family(fixtures_dir) -> dict:
    """Returns a LIMS-like case of samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(fixtures_dir, "report", "lims_family.json")
    )


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a LIMS case."""
    return lims_family["samples"]


@pytest.fixture(scope="function", name="mip_dna_analysis_hk_bundle_data")
def fixture_mip_dna_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, mip_dna_analysis_dir: Path, sample_id: str
) -> dict:
    """Return MIP DNA bundle data for Housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(mip_dna_analysis_dir, SNV_FILE_NAME).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "snv_research.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "sv_research.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "str.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "smn.vcf").as_posix(),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.mt.bam").as_posix(),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "vcf2cytosure.txt").as_posix(),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "multiqc.html").as_posix(),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
        ],
    }


@pytest.fixture(scope="function", name="mip_rna_analysis_hk_bundle_data")
def fixture_mip_rna_analysis_hk_bundle_data(
    rna_case_id: str,
    timestamp: datetime,
    mip_dna_analysis_dir: Path,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_mother_id: str,
    rna_sample_father_id: str,
) -> dict:
    """Return MIP RNA bundle data for Housekeeper."""

    files: [dict] = [
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.selected.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "clinical", rna_case_id],
        },
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "research", rna_case_id],
        },
    ]
    for sample_id in [
        rna_sample_son_id,
        rna_sample_daughter_id,
        rna_sample_mother_id,
        rna_sample_father_id,
    ]:
        files.extend(
            [
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1_star_sorted_sj.bigWig"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["coverage", "bigwig", "scout", sample_id],
                },
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1234_star_sorted_sj.bed.gz.tbi"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["bed", "scout", "junction", sample_id],
                },
            ]
        )

    return {
        "name": rna_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture(scope="function", name="balsamic_analysis_hk_bundle_data")
def fixture_balsamic_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, balsamic_wgs_analysis_dir: Path, sample_id: str
) -> dict:
    """Return Balsamic bundle data for Housekeeper,"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(balsamic_wgs_analysis_dir, SNV_FILE_NAME).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "umi.sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-umi-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "ascat.output.pdf").as_posix(),
                "archive": False,
                "tags": ["ascatngs", "visualization", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "coverage_qc_report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
        ],
    }


@pytest.fixture(scope="function", name="rnafusion_analysis_hk_bundle_data")
def fixture_rnafusion_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, rnafusion_analysis_dir: Path
) -> dict:
    """Get some bundle data for housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(rnafusion_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": ["multiqc-html", "rna"],
            },
            {
                "path": Path(rnafusion_analysis_dir, "rnafusion_report.html").as_posix(),
                "archive": False,
                "tags": ["fusionreport", "research"],
            },
        ],
    }


@pytest.fixture(name="balsamic_analysis_hk_version")
def fixture_balsamic_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a Balsamic bundle."""
    return helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_version")
def fixture_mip_dna_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a MIP DNA bundle."""
    return helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_api")
def fixture_mip_dna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP DNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_rna_analysis_hk_api")
def fixture_mip_rna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_rna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP RNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_rna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="balsamic_analysis_hk_api")
def fixture_balsamic_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with Balsamic analysis files."""
    helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="rnafusion_analysis_hk_api")
def fixture_rnafusion_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, rnafusion_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some rnafusion analysis files"""
    helpers.ensure_hk_version(housekeeper_api, rnafusion_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_file_handler")
def fixture_mip_file_handler(mip_dna_analysis_hk_version: Version) -> MipConfigBuilder:
    """Return a MIP confiig builder."""
    return MipConfigBuilder(hk_version_obj=mip_dna_analysis_hk_version)


@pytest.fixture(name="mip_dna_analysis")
def fixture_mip_dna_analysis(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers: StoreHelpers
) -> Analysis:
    """Return a MIP DNA analysis object."""
    helpers.add_synopsis_to_case(store=analysis_store_trio, case_id=case_id)
    case: Family = analysis_store_trio.family(case_id)
    analysis: Analysis = helpers.add_analysis(
        store=analysis_store_trio,
        case=case,
        started_at=timestamp,
        pipeline=Pipeline.MIP_DNA,
        completed_at=timestamp,
    )
    for link in case.links:
        helpers.add_phenotype_groups_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_phenotype_terms_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_subject_id_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
    return analysis


@pytest.fixture(name="balsamic_analysis_obj")
def fixture_balsamic_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic analysis object."""
    analysis_obj.pipeline = Pipeline.BALSAMIC
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.BALSAMIC
    return analysis_obj


@pytest.fixture(name="balsamic_umi_analysis_obj")
def fixture_balsamic_umi_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic UMI analysis object."""
    analysis_obj.pipeline = Pipeline.BALSAMIC_UMI
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.BALSAMIC_UMI

    return analysis_obj


@pytest.fixture(name="rnafusion_analysis_obj")
def fixture_rnafusion_analysis_obj(analysis_obj: models.Analysis) -> models.Analysis:
    """Return a RNAfusion analysis object."""
    analysis_obj.pipeline = Pipeline.RNAFUSION
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = (
            PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING
        )
        link_object.family.data_analysis = Pipeline.RNAFUSION
        link_object.family.panels = None
    return analysis_obj


@pytest.fixture(name="mip_config_builder")
def fixture_mip_config_builder(
    mip_dna_analysis_hk_version: Version,
    mip_dna_analysis: Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockMipAnalysis,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    """Return a MIP config builder."""
    return MipConfigBuilder(
        hk_version_obj=mip_dna_analysis_hk_version,
        analysis_obj=mip_dna_analysis,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )


@pytest.fixture(name="balsamic_config_builder")
def fixture_balsamic_config_builder(
    balsamic_analysis_hk_version: Version,
    balsamic_analysis_obj: Analysis,
    lims_api: MockLimsAPI,
) -> BalsamicConfigBuilder:
    """Return a Balsamic config builder."""
    return BalsamicConfigBuilder(
        hk_version_obj=balsamic_analysis_hk_version,
        analysis_obj=balsamic_analysis_obj,
        lims_api=lims_api,
    )


@pytest.fixture(name="mip_load_config")
def fixture_mip_load_config(
    mip_dna_analysis_dir: Path, case_id: str, customer_id: str
) -> MipLoadConfig:
    """Return a valid MIP load_config."""
    return MipLoadConfig(
        owner=customer_id,
        family=case_id,
        vcf_snv=Path(mip_dna_analysis_dir, SNV_FILE_NAME).as_posix(),
        track="rare",
    )


@pytest.fixture(name="lims_api")
def fixture_lims_api(lims_samples: List[dict]) -> MockLimsAPI:
    """Return a LIMS API."""
    return MockLimsAPI(samples=lims_samples)


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api() -> MockMipAnalysis:
    """Return a MIP analysis API."""
    return MockMipAnalysis()


@pytest.fixture(name="upload_scout_api")
def fixture_upload_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    housekeeper_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Return upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    return UploadScoutAPI(
        hk_api=housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_mip_analysis_scout_api")
def fixture_upload_mip_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return MIP upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=mip_dna_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_balsamic_analysis_scout_api")
def fixture_upload_balsamic_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    balsamic_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return Balsamic upload Scout API."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=balsamic_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_rnafusion_analysis_scout_api")
def fixture_upload_rnafusion_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    rnafusion_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api."""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=rnafusion_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )

    return _api



"""Tests for the compress fastq cli"""

import logging

from cg.cli.compress.fastq import (
    decompress_case,
    decompress_flowcell,
    decompress_sample,
    decompress_ticket,
)
from cg.models.cg_config import CGConfig
from click.testing import CliRunner


def test_decompress_spring_cli_no_family(
    compress_context: CGConfig, case_id: str, cli_runner: CliRunner, caplog
):
    """Test to run the decompress command when no families are found"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(decompress_case, ["nonexisting_case"], obj=compress_context)

    # THEN assert the program exits since no cases were found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert "Could not find case" in caplog.text


def test_decompress_spring_cli_one_family(
    populated_compress_context: CGConfig, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to run the decompress command with one case"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a case

    # WHEN running the compress command
    res = cli_runner.invoke(decompress_case, [case_id], obj=populated_compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Decompressed spring archives in 3 samples" in caplog.text



import datetime as dt
from pathlib import Path
from typing import Optional

from dateutil.parser import parse as parse_datestr
from pydantic import BaseModel, validator


class TrailblazerAnalysis(BaseModel):
    id: int
    family: str
    case_id = Optional[str]

    version: Optional[str]
    logged_at: Optional[str]
    started_at: Optional[str]
    completed_at: Optional[str]
    status: Optional[str]
    priority: Optional[str]
    out_dir: Optional[str]
    config_path: Optional[str]
    comment: Optional[str]
    is_deleted: Optional[bool]
    is_visible: Optional[bool]
    type: Optional[str]
    user_id: Optional[int]
    progress: Optional[float] = 0.0
    data_analysis: Optional[str]
    ticket: Optional[str]
    uploaded_at: Optional[str]

    @validator("case_id")
    def inherit_family_value(cls, value: str, values: dict) -> str:
        return values.get("family")

    @validator("logged_at", "started_at", "completed_at")
    def parse_str_to_datetime(cls, value: str) -> Optional[dt.datetime]:
        if value:
            return parse_datestr(value)

    @validator("out_dir", "config_path")
    def parse_str_to_path(cls, value: str) -> Optional[Path]:
        if value:
            return Path(value)

    class Config:
        extra = "allow"
        validate_all = True
        arbitrary_types_allowed = True



import datetime as dt
import re
from typing import List, Optional, Set, Dict

import alchy
from sqlalchemy import Column, ForeignKey, Table, UniqueConstraint, orm, types
from sqlalchemy.util import deprecated

from cg.constants import (
    CASE_ACTIONS,
    FLOWCELL_STATUS,
    PREP_CATEGORIES,
    Priority,
    SEX_OPTIONS,
    STATUS_OPTIONS,
    DataDelivery,
    Pipeline,
)

from cg.constants.constants import CONTROL_OPTIONS, PrepCategory

Model = alchy.make_declarative_base(Base=alchy.ModelBase)
flowcell_sample = Table(
    "flowcell_sample",
    Model.metadata,
    Column("flowcell_id", types.Integer, ForeignKey("flowcell.id"), nullable=False),
    Column("sample_id", types.Integer, ForeignKey("sample.id"), nullable=False),
    UniqueConstraint("flowcell_id", "sample_id", name="_flowcell_sample_uc"),
)

customer_user = Table(
    "customer_user",
    Model.metadata,
    Column("customer_id", types.Integer, ForeignKey("customer.id"), nullable=False),
    Column("user_id", types.Integer, ForeignKey("user.id"), nullable=False),
    UniqueConstraint("customer_id", "user_id", name="_customer_user_uc"),
)

customer_collaboration = Table(
    "customer_collaboration",
    Model.metadata,
    Column("customer_id", types.Integer, ForeignKey("customer.id"), nullable=False),
    Column("collaboration_id", types.Integer, ForeignKey("collaboration.id"), nullable=False),
    UniqueConstraint("customer_id", "collaboration_id", name="_customer_collaboration_uc"),
)


class PriorityMixin:
    @property
    def priority_human(self) -> str:
        """Humanized priority for sample."""
        return self.priority.name

    @priority_human.setter
    def priority_human(self, priority: str) -> None:
        self.priority: Priority = Priority[priority]

    @property
    def priority_int(self) -> int:
        """Priority as integer for sample."""
        return self.priority.value

    @priority_int.setter
    def priority_int(self, priority_int: int) -> None:
        self.priority: Priority = Priority(priority_int)

    @property
    def high_priority(self):
        """Has high priority?"""
        return self.priority_int > 1

    @property
    def low_priority(self):
        """Has low priority?"""
        return self.priority_int < 1


class Application(Model):
    id = Column(types.Integer, primary_key=True)
    tag = Column(types.String(32), unique=True, nullable=False)
    prep_category = Column(types.Enum(*PREP_CATEGORIES), nullable=False)
    is_external = Column(types.Boolean, nullable=False, default=False)
    description = Column(types.String(256), nullable=False)
    is_accredited = Column(types.Boolean, nullable=False)

    turnaround_time = Column(types.Integer)
    minimum_order = Column(types.Integer, default=1)
    sequencing_depth = Column(types.Integer)
    min_sequencing_depth = Column(types.Integer)
    target_reads = Column(types.BigInteger, default=0)
    percent_reads_guaranteed = Column(types.Integer, nullable=False)
    sample_amount = Column(types.Integer)
    sample_volume = Column(types.Text)
    sample_concentration = Column(types.Text)
    priority_processing = Column(types.Boolean, default=False)
    details = Column(types.Text)
    limitations = Column(types.Text)
    percent_kth = Column(types.Integer, nullable=False)
    comment = Column(types.Text)
    is_archived = Column(types.Boolean, default=False)

    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)
    versions = orm.relationship(
        "ApplicationVersion", order_by="ApplicationVersion.version", backref="application"
    )

    def __str__(self) -> str:
        return self.tag

    @property
    def reduced_price(self):
        return self.tag.startswith("WGT") or self.tag.startswith("EXT")

    @property
    def expected_reads(self):
        return self.target_reads * self.percent_reads_guaranteed / 100

    @property
    def analysis_type(self):
        if self.prep_category == PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING.value:
            return self.prep_category

        return (
            PrepCategory.WHOLE_GENOME_SEQUENCING.value
            if self.prep_category == PrepCategory.WHOLE_GENOME_SEQUENCING.value
            else PrepCategory.WHOLE_EXOME_SEQUENCING.value
        )


class ApplicationVersion(Model):
    __table_args__ = (UniqueConstraint("application_id", "version", name="_app_version_uc"),)

    id = Column(types.Integer, primary_key=True)
    version = Column(types.Integer, nullable=False)

    valid_from = Column(types.DateTime, default=dt.datetime.now, nullable=False)
    price_standard = Column(types.Integer)
    price_priority = Column(types.Integer)
    price_express = Column(types.Integer)
    price_research = Column(types.Integer)
    price_clinical_trials = Column(types.Integer)
    comment = Column(types.Text)

    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)
    application_id = Column(ForeignKey(Application.id), nullable=False)

    def __str__(self) -> str:
        return f"{self.application.tag} ({self.version})"

    def to_dict(self, application: bool = True):
        """Represent as dictionary"""
        data = super(ApplicationVersion, self).to_dict()
        if application:
            data["application"] = self.application.to_dict()
        return data


class Analysis(Model):
    id = Column(types.Integer, primary_key=True)
    pipeline = Column(types.Enum(*list(Pipeline)))
    pipeline_version = Column(types.String(32))
    started_at = Column(types.DateTime)
    completed_at = Column(types.DateTime)
    delivery_report_created_at = Column(types.DateTime)
    upload_started_at = Column(types.DateTime)
    uploaded_at = Column(types.DateTime)
    cleaned_at = Column(types.DateTime)
    # primary analysis is the one originally delivered to the customer
    is_primary = Column(types.Boolean, default=False)

    created_at = Column(types.DateTime, default=dt.datetime.now, nullable=False)
    family_id = Column(ForeignKey("family.id", ondelete="CASCADE"))
    uploaded_to_vogue_at = Column(types.DateTime, nullable=True)

    def __str__(self):
        return f"{self.family.internal_id} | {self.completed_at.date()}"

    def to_dict(self, family: bool = True):
        """Represent as dictionary"""
        data = super(Analysis, self).to_dict()
        if family:
            data["family"] = self.family.to_dict()
        return data


class Bed(Model):
    """Model for bed target captures"""

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(32), unique=True, nullable=False)
    comment = Column(types.Text)
    is_archived = Column(types.Boolean, default=False)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)

    versions = orm.relationship("BedVersion", order_by="BedVersion.version", backref="bed")

    def __str__(self) -> str:
        return self.name


class BedVersion(Model):
    """Model for bed target captures versions"""

    __table_args__ = (UniqueConstraint("bed_id", "version", name="_app_version_uc"),)

    id = Column(types.Integer, primary_key=True)
    shortname = Column(types.String(64))
    version = Column(types.Integer, nullable=False)
    filename = Column(types.String(256), nullable=False)
    checksum = Column(types.String(32))
    panel_size = Column(types.Integer)
    genome_version = Column(types.String(32))
    designer = Column(types.String(256))
    comment = Column(types.Text)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)
    bed_id = Column(ForeignKey(Bed.id), nullable=False)

    def __str__(self) -> str:
        return f"{self.bed.name} ({self.version})"

    def to_dict(self, bed: bool = True):
        """Represent as dictionary"""
        data = super(BedVersion, self).to_dict()
        if bed:
            data["bed"] = self.bed.to_dict()
        return data


class Customer(Model):
    agreement_date = Column(types.DateTime)
    agreement_registration = Column(types.String(32))
    comment = Column(types.Text)
    id = Column(types.Integer, primary_key=True)
    internal_id = Column(types.String(32), unique=True, nullable=False)
    invoice_address = Column(types.Text, nullable=False)
    invoice_reference = Column(types.String(32), nullable=False)
    loqus_upload = Column(types.Boolean, nullable=False, default=False)
    name = Column(types.String(128), nullable=False)
    organisation_number = Column(types.String(32))
    priority = Column(types.Enum("diagnostic", "research"))
    project_account_ki = Column(types.String(32))
    project_account_kth = Column(types.String(32))
    return_samples = Column(types.Boolean, nullable=False, default=False)
    scout_access = Column(types.Boolean, nullable=False, default=False)
    uppmax_account = Column(types.String(32))

    collaborations = orm.relationship("Collaboration", secondary=customer_collaboration)
    delivery_contact_id = Column(ForeignKey("user.id"))
    delivery_contact = orm.relationship("User", foreign_keys=[delivery_contact_id])
    invoice_contact_id = Column(ForeignKey("user.id"))
    invoice_contact = orm.relationship("User", foreign_keys=[invoice_contact_id])
    primary_contact_id = Column(ForeignKey("user.id"))
    primary_contact = orm.relationship("User", foreign_keys=[primary_contact_id])

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    def collaborators(self) -> Set["Customer"]:
        """All customers that the current customer collaborates with (including itself)"""
        customers = {
            customer
            for collaboration in self.collaborations
            for customer in collaboration.customers
        }
        customers.add(self)
        return customers


class Collaboration(Model):
    id = Column(types.Integer, primary_key=True)
    internal_id = Column(types.String(32), unique=True, nullable=False)
    name = Column(types.String(128), nullable=False)
    customers = orm.relationship(Customer, secondary=customer_collaboration)

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    def to_dict(self):
        """Represent as dictionary"""
        return {
            "customers": [customer.internal_id for customer in self.customers],
            "id": self.id,
            "name": self.name,
            "internal_id": self.internal_id,
        }


class Delivery(Model):
    id = Column(types.Integer, primary_key=True)
    delivered_at = Column(types.DateTime)
    removed_at = Column(types.DateTime)
    destination = Column(types.Enum("caesar", "pdc", "uppmax", "mh", "custom"), default="caesar")
    sample_id = Column(ForeignKey("sample.id", ondelete="CASCADE"))
    pool_id = Column(ForeignKey("pool.id", ondelete="CASCADE"))
    comment = Column(types.Text)


class Family(Model, PriorityMixin):
    __table_args__ = (UniqueConstraint("customer_id", "name", name="_customer_name_uc"),)

    action = Column(types.Enum(*CASE_ACTIONS))
    analyses = orm.relationship(Analysis, backref="family", order_by="-Analysis.completed_at")
    _cohorts = Column(types.Text)
    comment = Column(types.Text)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    customer_id = Column(ForeignKey("customer.id", ondelete="CASCADE"), nullable=False)
    customer = orm.relationship(Customer, foreign_keys=[customer_id])
    data_analysis = Column(types.Enum(*list(Pipeline)))
    data_delivery = Column(types.Enum(*list(DataDelivery)))
    id = Column(types.Integer, primary_key=True)
    internal_id = Column(types.String(32), unique=True, nullable=False)
    name = Column(types.String(128), nullable=False)
    ordered_at = Column(types.DateTime, default=dt.datetime.now)
    _panels = Column(types.Text)

    priority = Column(types.Enum(Priority), default=Priority.standard, nullable=False)
    synopsis = Column(types.Text)
    tickets = Column(types.VARCHAR)

    @property
    def cohorts(self) -> List[str]:
        """Return a list of cohorts."""
        return self._cohorts.split(",") if self._cohorts else []

    @cohorts.setter
    def cohorts(self, cohort_list: List[str]):
        self._cohorts = ",".join(cohort_list) if cohort_list else None

    @property
    def panels(self) -> List[str]:
        """Return a list of panels."""
        return self._panels.split(",") if self._panels else []

    @panels.setter
    def panels(self, panel_list: List[str]):
        self._panels = ",".join(panel_list) if panel_list else None

    @property
    def latest_ticket(self) -> Optional[str]:
        """Returns the last ticket the family was ordered in"""
        return self.tickets.split(sep=",")[-1] if self.tickets else None

    @property
    def latest_analyzed(self) -> Optional[dt.datetime]:
        return self.analyses[0].completed_at if self.analyses else None

    @property
    def latest_sequenced(self) -> Optional[dt.datetime]:
        sequenced_dates = []
        for link in self.links:
            if link.sample.application_version.application.is_external:
                sequenced_dates.append(link.sample.ordered_at)
            elif link.sample.sequenced_at:
                sequenced_dates.append(link.sample.sequenced_at)
        return max(sequenced_dates, default=None)

    @property
    def all_samples_pass_qc(self) -> bool:
        pass_qc = []
        for link in self.links:
            if link.sample.application_version.application.is_external or link.sample.sequencing_qc:
                pass_qc.append(True)
            else:
                pass_qc.append(False)
        return all(pass_qc)

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    def samples(self) -> List[str]:
        """Return case samples."""
        return self._get_samples

    @property
    def _get_samples(self) -> List[str]:
        """Extract samples from a case."""
        return [link.sample for link in self.links]

    @property
    def tumour_samples(self) -> List[str]:
        """Return tumour samples."""
        return self._get_tumour_samples

    @property
    def _get_tumour_samples(self) -> List[str]:
        """Extract tumour samples."""
        return [link.sample for link in self.links if link.sample.is_tumour]

    @property
    def loqusdb_uploaded_samples(self) -> List[str]:
        """Return uploaded samples to Loqusdb."""
        return self._get_loqusdb_uploaded_samples

    @property
    def _get_loqusdb_uploaded_samples(self) -> List[str]:
        """Extract samples uploaded to Loqusdb."""
        return [link.sample for link in self.links if link.sample.loqusdb_id]

    def get_delivery_arguments(self) -> Set[str]:
        """Translates the case data_delivery field to pipeline specific arguments."""
        delivery_arguments: Set[str] = set()
        requested_deliveries: List[str] = re.split("[-_]", self.data_delivery)
        delivery_per_pipeline_map: Dict[str, str] = {
            DataDelivery.FASTQ: Pipeline.FASTQ,
            DataDelivery.ANALYSIS_FILES: self.data_analysis,
        }
        for data_delivery, pipeline in delivery_per_pipeline_map.items():
            if data_delivery in requested_deliveries:
                delivery_arguments.add(pipeline)
        return delivery_arguments

    def to_dict(self, links: bool = False, analyses: bool = False) -> dict:
        """Represent as dictionary."""
        data = super(Family, self).to_dict()
        data["panels"] = self.panels
        data["priority"] = self.priority_human
        data["customer"] = self.customer.to_dict()
        if links:
            data["links"] = [link_obj.to_dict(samples=True) for link_obj in self.links]
        if analyses:
            data["analyses"] = [
                analysis_obj.to_dict(family=False) for analysis_obj in self.analyses
            ]
        return data


class FamilySample(Model):
    __table_args__ = (UniqueConstraint("family_id", "sample_id", name="_family_sample_uc"),)

    id = Column(types.Integer, primary_key=True)
    family_id = Column(ForeignKey("family.id", ondelete="CASCADE"), nullable=False)
    sample_id = Column(ForeignKey("sample.id", ondelete="CASCADE"), nullable=False)
    status = Column(types.Enum(*STATUS_OPTIONS), default="unknown", nullable=False)

    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)

    mother_id = Column(ForeignKey("sample.id"))
    father_id = Column(ForeignKey("sample.id"))

    family = orm.relationship("Family", backref="links")
    sample = orm.relationship("Sample", foreign_keys=[sample_id], backref="links")
    mother = orm.relationship("Sample", foreign_keys=[mother_id], backref="mother_links")
    father = orm.relationship("Sample", foreign_keys=[father_id], backref="father_links")

    def to_dict(self, parents: bool = False, samples: bool = False, family: bool = False) -> dict:
        """Represent as dictionary"""
        data = super(FamilySample, self).to_dict()
        if samples:
            data["sample"] = self.sample.to_dict()
            data["mother"] = self.mother.to_dict() if self.mother else None
            data["father"] = self.father.to_dict() if self.father else None
        elif parents:
            data["mother"] = self.mother.to_dict() if self.mother else None
            data["father"] = self.father.to_dict() if self.father else None
        if family:
            data["family"] = self.family.to_dict()
        return data

    def __str__(self) -> str:
        return f"{self.family.internal_id} | {self.sample.internal_id}"


class Flowcell(Model):
    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(32), unique=True, nullable=False)
    sequencer_type = Column(types.Enum("hiseqga", "hiseqx", "novaseq"))
    sequencer_name = Column(types.String(32))
    sequenced_at = Column(types.DateTime)
    status = Column(types.Enum(*FLOWCELL_STATUS), default="ondisk")
    archived_at = Column(types.DateTime)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)

    samples = orm.relationship("Sample", secondary=flowcell_sample, backref="flowcells")

    def __str__(self):
        return self.name

    def to_dict(self, samples: bool = False):
        """Represent as dictionary"""
        data = super(Flowcell, self).to_dict()
        if samples:
            data["samples"] = [sample.to_dict() for sample in self.samples]
        return data


class Organism(Model):
    id = Column(types.Integer, primary_key=True)
    internal_id = Column(types.String(32), nullable=False, unique=True)
    name = Column(types.String(255), nullable=False, unique=True)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)
    reference_genome = Column(types.String(255))
    verified = Column(types.Boolean, default=False)
    comment = Column(types.Text)

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    def to_dict(self) -> dict:
        """Represent as dictionary"""
        return super(Organism, self).to_dict()


class Panel(Model):
    abbrev = Column(types.String(32), unique=True)
    current_version = Column(types.Float, nullable=False)
    customer_id = Column(ForeignKey("customer.id", ondelete="CASCADE"), nullable=False)
    customer = orm.relationship(Customer, backref="panels")
    date = Column(types.DateTime, nullable=False)
    gene_count = Column(types.Integer)
    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(64), unique=True)

    def __str__(self):
        return f"{self.abbrev} ({self.current_version})"


class Pool(Model):
    __table_args__ = (UniqueConstraint("order", "name", name="_order_name_uc"),)

    application_version_id = Column(ForeignKey("application_version.id"), nullable=False)
    application_version = orm.relationship(
        ApplicationVersion, foreign_keys=[application_version_id]
    )
    comment = Column(types.Text)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    customer_id = Column(ForeignKey("customer.id", ondelete="CASCADE"), nullable=False)
    customer = orm.relationship(Customer, foreign_keys=[customer_id])
    delivered_at = Column(types.DateTime)
    deliveries = orm.relationship(Delivery, backref="pool")
    id = Column(types.Integer, primary_key=True)
    invoice_id = Column(ForeignKey("invoice.id"))
    name = Column(types.String(32), nullable=False)
    no_invoice = Column(types.Boolean, default=False)
    order = Column(types.String(64), nullable=False)
    ordered_at = Column(types.DateTime, nullable=False)
    received_at = Column(types.DateTime)
    ticket = Column(types.String(32))


class Sample(Model, PriorityMixin):
    age_at_sampling = Column(types.FLOAT)
    application_version_id = Column(ForeignKey("application_version.id"), nullable=False)
    application_version = orm.relationship(
        ApplicationVersion, foreign_keys=[application_version_id]
    )
    capture_kit = Column(types.String(64))
    comment = Column(types.Text)
    control = Column(types.Enum(*CONTROL_OPTIONS))
    created_at = Column(types.DateTime, default=dt.datetime.now)
    customer_id = Column(ForeignKey("customer.id", ondelete="CASCADE"), nullable=False)
    customer = orm.relationship("Customer", foreign_keys=[customer_id])
    delivered_at = Column(types.DateTime)
    deliveries = orm.relationship(Delivery, backref="sample")
    downsampled_to = Column(types.BigInteger)
    from_sample = Column(types.String(128))
    id = Column(types.Integer, primary_key=True)
    internal_id = Column(types.String(32), nullable=False, unique=True)
    invoice_id = Column(ForeignKey("invoice.id"))
    invoiced_at = Column(types.DateTime)  # DEPRECATED
    _is_external = Column("is_external", types.Boolean)  # DEPRECATED
    is_tumour = Column(types.Boolean, default=False)
    loqusdb_id = Column(types.String(64))
    name = Column(types.String(128), nullable=False)
    no_invoice = Column(types.Boolean, default=False)
    order = Column(types.String(64))
    ordered_at = Column(types.DateTime, nullable=False)
    organism_id = Column(ForeignKey("organism.id"))
    organism = orm.relationship("Organism", foreign_keys=[organism_id])
    original_ticket = Column(types.String(32))
    _phenotype_groups = Column(types.Text)
    _phenotype_terms = Column(types.Text)
    prepared_at = Column(types.DateTime)

    priority = Column(types.Enum(Priority), default=Priority.standard, nullable=False)
    reads = Column(types.BigInteger, default=0)
    received_at = Column(types.DateTime)
    reference_genome = Column(types.String(255))
    sequence_start = Column(types.DateTime)
    sequenced_at = Column(types.DateTime)
    sex = Column(types.Enum(*SEX_OPTIONS), nullable=False)
    subject_id = Column(types.String(128))

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    @deprecated(
        version="1.4.0",
        message="This field is deprecated, use sample.application_version.application.is_external",
    )
    def is_external(self):
        """Return if this is an externally sequenced sample."""
        return self._is_external

    @property
    def sequencing_qc(self) -> bool:
        """Return sequencing qc passed or failed."""
        application = self.application_version.application
        # Express priority needs to be analyzed at a lower threshold for primary analysis
        if self.priority == Priority.express:
            one_half_of_target_reads = application.target_reads / 2
            return self.reads >= one_half_of_target_reads
        if self.application_version.application.prep_category == PrepCategory.READY_MADE_LIBRARY:
            return bool(self.reads)
        return self.reads > application.expected_reads

    @property
    def phenotype_groups(self) -> List[str]:
        """Return a list of phenotype_groups."""
        return self._phenotype_groups.split(",") if self._phenotype_groups else []

    @phenotype_groups.setter
    def phenotype_groups(self, phenotype_term_list: List[str]):
        self._phenotype_groups = ",".join(phenotype_term_list) if phenotype_term_list else None

    @property
    def phenotype_terms(self) -> List[str]:
        """Return a list of phenotype_terms."""
        return self._phenotype_terms.split(",") if self._phenotype_terms else []

    @phenotype_terms.setter
    def phenotype_terms(self, phenotype_term_list: List[str]):
        self._phenotype_terms = ",".join(phenotype_term_list) if phenotype_term_list else None

    @property
    def state(self) -> str:
        """Get the current sample state."""
        if self.delivered_at:
            return f"Delivered {self.delivered_at.date()}"
        if self.sequenced_at:
            return f"Sequenced {self.sequenced_at.date()}"
        if self.sequence_start:
            return f"Sequencing {self.sequence_start.date()}"
        if self.received_at:
            return f"Received {self.received_at.date()}"

        return f"Ordered {self.ordered_at.date()}"

    def to_dict(self, links: bool = False, flowcells: bool = False) -> dict:
        """Represent as dictionary"""
        data = super(Sample, self).to_dict()
        data["priority"] = self.priority_human
        data["customer"] = self.customer.to_dict()
        data["application_version"] = self.application_version.to_dict()
        data["application"] = self.application_version.application.to_dict()
        if links:
            data["links"] = [link_obj.to_dict(family=True, parents=True) for link_obj in self.links]
        if flowcells:
            data["flowcells"] = [flowcell_obj.to_dict() for flowcell_obj in self.flowcells]
        return data


class Invoice(Model):
    id = Column(types.Integer, primary_key=True)
    customer_id = Column(ForeignKey("customer.id"), nullable=False)
    customer = orm.relationship(Customer, foreign_keys=[customer_id])
    created_at = Column(types.DateTime, default=dt.datetime.now)
    updated_at = Column(types.DateTime, onupdate=dt.datetime.now)
    invoiced_at = Column(types.DateTime)
    comment = Column(types.Text)
    discount = Column(types.Integer, default=0)
    excel_kth = Column(types.BLOB)
    excel_ki = Column(types.BLOB)
    price = Column(types.Integer)
    record_type = Column(types.Text)

    samples = orm.relationship(Sample, backref="invoice")
    pools = orm.relationship(Pool, backref="invoice")

    def __str__(self):
        return f"{self.customer_id} ({self.invoiced_at})"

    def to_dict(self) -> dict:
        """Represent as dictionary"""
        return super(Invoice, self).to_dict()


class User(Model):
    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(128), nullable=False)
    email = Column(types.String(128), unique=True, nullable=False)
    is_admin = Column(types.Boolean, default=False)
    order_portal_login = Column(types.Boolean, default=False)

    customers = orm.relationship("Customer", secondary=customer_user, backref="users")

    def to_dict(self) -> dict:
        """Represent as dictionary"""
        data = super(User, self).to_dict()
        data["customers"] = [record.to_dict() for record in self.customers]
        return data

    def __str__(self) -> str:
        return self.name



"""Helper functions."""

from typing import Optional, List


def get_string_from_list_by_pattern(strings: List[str], pattern: str) -> Optional[str]:
    """Returns the full string from a list given a specific pattern."""

    return next((string for string in strings if pattern in string), None) if strings else None



"""Fixtures for crunchy API."""
import logging
from pathlib import Path
from typing import List, Dict

import pytest

from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.models import CompressionData
from cgmodels.crunchy.metadata import CrunchyMetadata

LOG = logging.getLogger(__name__)


@pytest.fixture(name="real_spring_metadata_path")
def fixture_real_spring_metadata_path(apps_dir: Path) -> Path:
    """Return the path to a SPRING metadata file."""
    return Path(apps_dir, "crunchy", "spring_metadata.json")


@pytest.fixture(name="spring_metadata")
def fixture_spring_metadata(compression_object: CompressionData) -> List[dict]:
    """Return meta data information."""
    return [
        {
            "path": compression_object.fastq_first.as_posix(),
            "file": "first_read",
            "checksum": "checksum_first_read",
            "algorithm": "sha256",
        },
        {
            "path": compression_object.fastq_second.as_posix(),
            "file": "second_read",
            "checksum": "checksum_second_read",
            "algorithm": "sha256",
        },
        {"path": compression_object.spring_path.as_posix(), "file": "spring"},
    ]


@pytest.fixture(name="crunchy_metadata_object")
def fixture_crunchy_metadata_object(spring_metadata: List[dict]) -> CrunchyMetadata:
    """Return Crunchy metadata."""
    return CrunchyMetadata(files=spring_metadata)


@pytest.fixture(name="spring_metadata_file")
def fixture_spring_metadata_file(
    compression_object: CompressionData, spring_metadata: List[dict]
) -> Path:
    """Return the path to a populated SPRING metadata file."""
    metadata_path = compression_object.spring_metadata_path
    WriteFile.write_file_from_content(
        content=spring_metadata, file_format=FileFormat.JSON, file_path=metadata_path
    )
    return metadata_path


@pytest.fixture(name="fastq_first_file")
def fixture_fastq_first_file(fastq_first_path: Path) -> Path:
    """Creates an existing FASTQ path."""
    fastq_first_path.touch()
    return fastq_first_path


@pytest.fixture(name="fastq_second_file")
def fixture_fastq_second_file(fastq_second_path: Path) -> Path:
    """Creates an existing FASTQ path."""
    fastq_second_path.touch()
    return fastq_second_path


@pytest.fixture(name="spring_file")
def fixture_spring_file(spring_path: Path) -> Path:
    """Creates an existing SPRING file."""
    spring_path.touch()
    return spring_path


@pytest.fixture(name="fastq_paths")
def fixture_fastq_paths(fastq_first_path: Path, fastq_second_path: Path) -> Dict[str, Path]:
    """Creates FASTQ paths."""
    return {
        "fastq_first_path": fastq_first_path,
        "fastq_second_path": fastq_second_path,
    }


@pytest.fixture(name="existing_fastq_paths")
def fixture_existing_fastq_paths(
    fastq_first_file: Path, fastq_second_file: Path
) -> Dict[str, Path]:
    """Creates existing FASTQ paths."""
    return {
        "fastq_first_path": fastq_first_file,
        "fastq_second_path": fastq_second_file,
    }



""" Module for retrieving flow cells from backup."""
import logging
import re
import subprocess
from pathlib import Path
from typing import Optional, Dict, List, Tuple

from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.constants import FileExtensions, FlowCellStatus
from cg.constants.backup import MAX_PROCESSING_FLOW_CELLS
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.indexes import ListIndexes
from cg.constants.pdc import PDCExitCodes
from cg.constants.process import RETURN_WARNING
from cg.constants.symbols import ASTERISK, NEW_LINE
from cg.exc import ChecksumFailedError, PdcNoFilesMatchingSearchError
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import EncryptionAPI, SpringEncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models import CompressionData
from cg.store import Store
from cg.store.models import Flowcell
from cg.utils.time import get_elapsed_time, get_start_time

LOG = logging.getLogger(__name__)


class BackupAPI:
    """Class for retrieving FCs from backup."""

    def __init__(
        self,
        encryption_api: EncryptionAPI,
        encrypt_dir: Dict[str, str],
        status: Store,
        tar_api: TarAPI,
        pdc_api: PdcAPI,
        root_dir: Dict[str, str],
        dry_run: bool = False,
    ):
        self.encryption_api = encryption_api
        self.encrypt_dir = encrypt_dir
        self.status: Store = status
        self.tar_api: TarAPI = tar_api
        self.pdc: PdcAPI = pdc_api
        self.root_dir: dict = root_dir
        self.dry_run: bool = dry_run

    def check_processing(self) -> bool:
        """Check if the processing queue for flow cells is not full."""
        processing_flow_cells_count: int = self.status.get_flow_cells_by_statuses(
            flow_cell_statuses=[FlowCellStatus.PROCESSING]
        ).count()
        LOG.debug(f"Processing flow cells: {processing_flow_cells_count}")
        return processing_flow_cells_count < MAX_PROCESSING_FLOW_CELLS

    def get_first_flow_cell(self) -> Optional[Flowcell]:
        """Get the first flow cell from the requested queue."""
        flow_cell: Optional[Flowcell] = self.status.get_flow_cells_by_statuses(
            flow_cell_statuses=[FlowCellStatus.REQUESTED]
        ).first()
        return flow_cell or None

    def fetch_flow_cell(self, flow_cell: Optional[Flowcell] = None) -> Optional[float]:
        """Start fetching a flow cell from backup if possible.

        1. The processing queue is not full.
        2. The requested queue is not emtpy.
        """
        if self.check_processing() is False:
            LOG.info("Processing queue is full")
            return None

        if not flow_cell:
            flow_cell: Optional[Flowcell] = self.get_first_flow_cell()

        if not flow_cell:
            LOG.info("No flow cells requested")
            return None

        flow_cell.status: FlowCellStatus = FlowCellStatus.PROCESSING
        if not self.dry_run:
            self.status.commit()
            LOG.info(f"{flow_cell.name}: retrieving from PDC")

        try:
            pdc_flow_cell_query: List[str] = self.query_pdc_for_flow_cell(flow_cell.name)

        except PdcNoFilesMatchingSearchError as error:
            LOG.error(f"PDC query failed: {error}")
            raise error

        archived_key: Path = self.get_archived_encryption_key_path(query=pdc_flow_cell_query)
        archived_flow_cell: Path = self.get_archived_flow_cell_path(query=pdc_flow_cell_query)

        if not self.dry_run:
            return self._process_flow_cell(
                flow_cell=flow_cell,
                archived_key=archived_key,
                archived_flow_cell=archived_flow_cell,
            )

    def _process_flow_cell(
        self, flow_cell: Flowcell, archived_key: Path, archived_flow_cell: Path
    ) -> float:
        """Process a flow cell from backup. Return elapsed time."""
        start_time: float = get_start_time()
        run_dir: Path = Path(self.root_dir[flow_cell.sequencer_type])
        self.retrieve_archived_key(archived_key=archived_key, flow_cell=flow_cell, run_dir=run_dir)
        self.retrieve_archived_flow_cell(
            archived_flow_cell=archived_flow_cell, flow_cell=flow_cell, run_dir=run_dir
        )

        try:
            (
                decrypted_flow_cell,
                encryption_key,
                retrieved_flow_cell,
                retrieved_key,
            ) = self.decrypt_flow_cell(archived_flow_cell, archived_key, run_dir)

            self.extract_flow_cell(decrypted_flow_cell, run_dir)
            self.create_rta_complete(decrypted_flow_cell, run_dir)
            self.unlink_files(
                decrypted_flow_cell, encryption_key, retrieved_flow_cell, retrieved_key
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            if not self.dry_run:
                flow_cell.status: str = FlowCellStatus.REQUESTED
                self.status.commit()
            raise error

        return get_elapsed_time(start_time=start_time)

    def unlink_files(
        self,
        decrypted_flow_cell: Path,
        encryption_key: Path,
        retrieved_flow_cell: Path,
        retrieved_key: Path,
    ):
        """Remove files after flow cell has been fetched from PDC."""
        if self.dry_run:
            return
        LOG.debug("Unlink files")
        message = f"{retrieved_flow_cell} not found, skipping removal"
        try:
            retrieved_flow_cell.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            decrypted_flow_cell.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            retrieved_key.unlink()
        except FileNotFoundError:
            LOG.info(message)
        try:
            encryption_key.unlink()
        except FileNotFoundError:
            LOG.info(message)

    @staticmethod
    def create_rta_complete(decrypted_flow_cell: Path, run_dir: Path):
        """Create an RTAComplete.txt file in the flow cell run directory."""
        (
            run_dir / Path(decrypted_flow_cell.stem).stem / DemultiplexingDirsAndFiles.RTACOMPLETE
        ).touch()

    def extract_flow_cell(self, decrypted_flow_cell, run_dir):
        """Extract the flow cell tar archive."""
        extraction_command = self.tar_api.get_extract_file_command(
            input_file=decrypted_flow_cell, output_dir=run_dir
        )
        LOG.debug(f"Extract flow cell command: {extraction_command}")
        self.tar_api.run_tar_command(extraction_command)

    def decrypt_flow_cell(
        self, archived_flow_cell: Path, archived_key: Path, run_dir: Path
    ) -> Tuple[Path, Path, Path, Path]:
        """Decrypt the flow cell."""
        retrieved_key: Path = run_dir / archived_key.name
        encryption_key: Path = retrieved_key.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: List[str] = self.encryption_api.get_asymmetric_decryption_command(
            input_file=retrieved_key, output_file=encryption_key
        )
        LOG.debug(f"Decrypt key command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        retrieved_flow_cell: Path = run_dir / archived_flow_cell.name
        decrypted_flow_cell: Path = retrieved_flow_cell.with_suffix(FileExtensions.NO_EXTENSION)
        decryption_command: List[str] = self.encryption_api.get_symmetric_decryption_command(
            input_file=retrieved_flow_cell,
            output_file=decrypted_flow_cell,
            encryption_key=encryption_key,
        )
        LOG.debug(f"Decrypt flow cell command: {decryption_command}")
        self.encryption_api.run_gpg_command(decryption_command)
        return decrypted_flow_cell, encryption_key, retrieved_flow_cell, retrieved_key

    def retrieve_archived_key(self, archived_key: Path, flow_cell: Flowcell, run_dir: Path) -> None:
        """Attempt to retrieve an archived key."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_key,
                run_dir=run_dir,
            )
        except subprocess.CalledProcessError as error:
            if error.returncode == RETURN_WARNING:
                LOG.warning(
                    f"WARNING for retrieval of encryption key of flow cell {flow_cell.name}, please check "
                    "dsmerror.log"
                )
            else:
                LOG.error(f"{flow_cell.name}: key retrieval failed")
                if not self.dry_run:
                    flow_cell.status = FlowCellStatus.REQUESTED
                    self.status.commit()
                raise error

    def retrieve_archived_flow_cell(
        self, archived_flow_cell: Path, flow_cell: Flowcell, run_dir: Path
    ):
        """Attempt to retrieve an archived flow cell."""
        try:
            self.retrieve_archived_file(
                archived_file=archived_flow_cell,
                run_dir=run_dir,
            )
            if not self.dry_run:
                self._set_flow_cell_status_to_retrieved(flow_cell)
        except subprocess.CalledProcessError as error:
            if error.returncode == RETURN_WARNING:
                LOG.warning(
                    f"WARNING for retrieval of flow cell {flow_cell.name}, please check dsmerror.log"
                )
                if not self.dry_run:
                    self._set_flow_cell_status_to_retrieved(flow_cell)
            else:
                LOG.error(f"{flow_cell.name}: run directory retrieval failed")
                if not self.dry_run:
                    flow_cell.status = FlowCellStatus.REQUESTED
                    self.status.commit()
                raise error

    def _set_flow_cell_status_to_retrieved(self, flow_cell: Flowcell):
        flow_cell.status = FlowCellStatus.RETRIEVED
        self.status.commit()
        LOG.info(f"Status for flow cell {flow_cell.name} set to {flow_cell.status}")

    def query_pdc_for_flow_cell(self, flow_cell_id) -> List[str]:
        """Query PDC for a given flow cell id."""
        search_patterns: List[str] = [
            dir + ASTERISK + flow_cell_id + ASTERISK for dir in self.encrypt_dir.values()
        ]

        for search_pattern in search_patterns:
            try:
                self.pdc.query_pdc(search_pattern=search_pattern)
                query: List[str] = self.pdc.process.stdout.split(NEW_LINE)
                continue
            except subprocess.CalledProcessError as error:
                if error.returncode != PDCExitCodes.NO_FILES_FOUND:
                    raise error
                LOG.info(
                    f"No archived files found for pdc query {search_patterns}, testing legacy directory"
                )

        if not query:
            raise PdcNoFilesMatchingSearchError(
                message=f"No archived files found for pdc queries {search_patterns}"
            )
        LOG.info(f"Found archived files for pdc queries {search_patterns}")
        return query

    def retrieve_archived_file(self, archived_file: Path, run_dir: Path) -> None:
        """Retrieve the archived file from PDC to a flow cell runs directory."""
        retrieved_file: Path = run_dir / archived_file.name
        LOG.debug(f"Retrieving file {archived_file} to {retrieved_file}")
        self.pdc.retrieve_file_from_pdc(
            file_path=str(archived_file), target_path=str(retrieved_file)
        )

    def get_archived_flow_cell_path(self, query: list) -> Path:
        """Get the path of the archived flow cell from a PDC query."""
        flow_cell_query: str = [
            row
            for row in query
            if FileExtensions.TAR in row
            and FileExtensions.GZIP in row
            and FileExtensions.GPG in row
        ][ListIndexes.FIRST.value]

        for dir in self.encrypt_dir.values():
            re_archived_flow_cell_path: re.Pattern = re.compile(dir + ".+?(?=\s)")
            arhived_flow_cell: Optional[re.Match] = re.search(
                re_archived_flow_cell_path, flow_cell_query
            )
            if arhived_flow_cell:
                archived_flow_cell_path: Path = Path(arhived_flow_cell.group())
                LOG.info(f"Flow cell found: {archived_flow_cell_path}")
                return archived_flow_cell_path

    def get_archived_encryption_key_path(self, query: list) -> Path:
        """Get the encryption key for the archived flow cell from a PDC query."""
        encryption_key_query: str = [
            row for row in query if FileExtensions.KEY in row and FileExtensions.GPG in row
        ][ListIndexes.FIRST.value]

        for dir in self.encrypt_dir.values():
            re_archived_encryption_key_path: re.Pattern = re.compile(dir + ".+?(?=\s)")
            archived_encryption_key: Optional[re.Match] = re.search(
                re_archived_encryption_key_path, encryption_key_query
            )
            if archived_encryption_key:
                archived_encryption_key_path: Path = Path(archived_encryption_key.group())
                LOG.info(f"Encryption key found: {archived_encryption_key_path}")
                return archived_encryption_key_path


class SpringBackupAPI:
    """Class for handling PDC backup and retrieval of spring compressed fastq files."""

    def __init__(
        self,
        encryption_api: SpringEncryptionAPI,
        pdc_api: PdcAPI,
        hk_api: HousekeeperAPI,
        dry_run: bool = False,
    ):
        self.encryption_api: SpringEncryptionAPI = encryption_api
        self.hk_api: HousekeeperAPI = hk_api
        self.pdc: PdcAPI = pdc_api
        self.dry_run = dry_run

    def encrypt_and_archive_spring_file(self, spring_file_path: Path) -> None:
        """Encrypts and archives a spring file and its decryption key."""
        LOG.debug(f"*** START BACKUP PROCESS OF SPRING FILE {spring_file_path} ***")
        if self.is_compression_ongoing(spring_file_path):
            LOG.info(
                f"Spring (de)compression ongoing, skipping archiving for spring file {spring_file_path}",
            )
            return
        self.encryption_api.cleanup(spring_file_path)
        if self.is_spring_file_archived(spring_file_path):
            LOG.info(
                "Spring file already archived! Spring file will be removed from disk, continuing "
                "to next spring file "
            )
            self.remove_archived_spring_file(spring_file_path)
            return
        try:
            self.encryption_api.spring_symmetric_encryption(spring_file_path)
            self.encryption_api.key_asymmetric_encryption(spring_file_path)
            self.encryption_api.compare_spring_file_checksums(spring_file_path)
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.mark_file_as_archived(spring_file_path)
            self.encryption_api.cleanup(spring_file_path)
            self.remove_archived_spring_file(spring_file_path)
            LOG.debug("*** ARCHIVING PROCESS COMPLETED SUCCESSFULLY ***")
        except subprocess.CalledProcessError as error:
            LOG.error(f"Encryption failed: {error.stderr}")
            LOG.debug("*** COMMAND PROCESS FAILED! ***")
            self.encryption_api.cleanup(spring_file_path)
        except ChecksumFailedError as error:
            LOG.error(error)
            self.encryption_api.cleanup(spring_file_path)
            LOG.debug("*** CHECKSUM PROCESS FAILED! ***")

    def retrieve_and_decrypt_spring_file(self, spring_file_path: Path) -> None:
        """Retrieves and decrypts a spring file and its decryption key."""
        LOG.info(f"*** START RETRIEVAL PROCESS OF SPRING FILE {spring_file_path} ***")
        try:
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
            )
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
            )
            self.encryption_api.key_asymmetric_decryption(spring_file_path)
            self.encryption_api.spring_symmetric_decryption(
                spring_file_path, output_file=spring_file_path
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            LOG.debug("*** RETRIEVAL PROCESS FAILED! ***")
        self.encryption_api.cleanup(spring_file_path)
        LOG.debug("*** RETRIEVAL PROCESS COMPLETED SUCCESSFULLY ***")

    def mark_file_as_archived(self, spring_file_path: Path) -> None:
        """Set the field 'to_archive' of the file in Housekeeper to mark that it has been
        archived to PDC."""
        if self.dry_run:
            LOG.info(f"Dry run, no changes made to {spring_file_path}")
            return
        hk_spring_file: hk_models.File = self.hk_api.files(path=str(spring_file_path)).first()
        LOG.info(f"Setting {spring_file_path} to archived in Housekeeper")
        self.hk_api.set_to_archive(file=hk_spring_file, value=True)

    def is_to_be_retrieved_and_decrypted(self, spring_file_path: Path) -> bool:
        """Determines if a spring file is archived on PDC and needs to be retrieved and decrypted."""
        return (
            self.hk_api.files(path=str(spring_file_path)).first().to_archive
            and not spring_file_path.exists()
        )

    def remove_archived_spring_file(self, spring_file_path: Path) -> None:
        """Removes all files related to spring PDC archiving."""
        if not self.dry_run:
            LOG.info(f"Removing spring file {spring_file_path} from disk")
            spring_file_path.unlink()

    def is_spring_file_archived(self, spring_file_path: Path) -> bool:
        """Checks if a spring file is marked as archived in Housekeeper."""
        return self.hk_api.files(path=str(spring_file_path)).first().to_archive

    @staticmethod
    def is_compression_ongoing(spring_file_path: Path) -> bool:
        """Determines if (de)compression of the spring file ongoing."""
        return CompressionData(spring_file_path).pending_exists()



"""Tests API upload methods for balsamic cases"""

from cg.meta.upload.gt import UploadGenotypesAPI
from cg.models.cg_config import CGConfig
from cg.store import models
from tests.cli.workflow.balsamic.conftest import (
    fixture_balsamic_context,
    balsamic_lims,
    balsamic_housekeeper,
    balsamic_mock_fastq_files,
    balsamic_housekeeper_dir,
    fastq_file_l_1_r_1,
    fastq_file_l_1_r_2,
    fastq_file_l_2_r_1,
    fastq_file_l_2_r_2,
    fastq_file_l_3_r_1,
    fastq_file_l_3_r_2,
    fastq_file_l_4_r_1,
    fastq_file_l_4_r_2,
)


def test_genotype_check_wgs_normal(balsamic_context: CGConfig):
    """Test a cancer case with WGS and normal sample that is Genotype compatible."""
    # GIVEN a balsamic case with WGS tag and a normal sample
    internal_id = "balsamic_case_wgs_paired_enough_reads"
    case_obj: models.Family = balsamic_context.status_db.family(internal_id=internal_id)

    # WHEN checking if the case is Genotype upload compatible
    passed_check = UploadGenotypesAPI.is_suitable_for_genotype_upload(case_obj)

    # THEN it should return True
    assert passed_check


def test_genotype_check_non_wgs_normal(balsamic_context: CGConfig):
    """Test a cancer case with no WGS sample that is not Genotype compatible."""
    # GIVEN a balsamic case with a normal sample, but no WGS tag
    internal_id = "balsamic_case_tgs_paired"
    case_obj: models.Family = balsamic_context.status_db.family(internal_id=internal_id)

    # WHEN checking if the case is Genotype upload compatible
    passed_check = UploadGenotypesAPI.is_suitable_for_genotype_upload(case_obj)

    # THEN it should return False
    assert not passed_check


def test_genotype_check_only_tumour(balsamic_context: CGConfig):
    """Test a cancer case with only a tumour sample that is not Genotype compatible."""
    # GIVEN a balsamic case with only tumour sample
    internal_id = "balsamic_case_wgs_single"
    case_obj: models.Family = balsamic_context.status_db.family(internal_id=internal_id)

    # WHEN checking if the case is Genotype upload compatible
    passed_check = UploadGenotypesAPI.is_suitable_for_genotype_upload(case_obj)

    # THEN it should return False
    assert not passed_check



from typing import List

from cg.store import Store
from cg.store.models import Bed
from cg.store.status_bed_filters import get_not_archived_beds, order_beds_by_name


def test_get_beds_by_name(base_store: Store, bed_name: str):
    """Test return bed by name."""
    # GIVEN a store containing bed

    # WHEN retrieving a beds by name
    beds: List[Bed] = get_not_archived_beds(beds=base_store._get_bed_query())

    # THEN beds should be returned
    assert beds

    # THEN the bed attribute name should match the original
    assert beds[0].name == bed_name


def test_get_not_archived_beds(base_store: Store):
    """Test return bed not archived."""
    # GIVEN a store containing bed

    # WHEN retrieving a beds not archived
    beds: List[Bed] = get_not_archived_beds(beds=base_store._get_bed_query())

    # THEN beds should be returned
    assert beds

    # THEN the bed attribute is_archived should be False
    assert beds[0].is_archived is False


def test_order_beds_by_name(base_store: Store, bed_name: str):
    """Test return bed ordered by name."""
    # GIVEN a store containing bed

    # WHEN retrieving a beds
    beds: List[Bed] = order_beds_by_name(beds=base_store._get_bed_query())

    # THEN a bed should be returned
    assert beds

    # THEN the bed attribute name should match the original
    assert beds[0].name == bed_name



import http
import logging
import json
import tempfile
from functools import wraps
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
from sqlalchemy.exc import IntegrityError
from urllib3.exceptions import MaxRetryError, NewConnectionError

from cg.apps.orderform.excel_orderform_parser import ExcelOrderformParser
from cg.apps.orderform.json_orderform_parser import JsonOrderformParser
from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.constants import FileFormat
from cg.exc import OrderError, OrderFormError, TicketCreationError
from cg.server.ext import db, lims, osticket
from cg.io.controller import WriteStream
from cg.meta.orders import OrdersAPI
from cg.store.models import Customer
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.orderform_schema import Orderform
from flask import Blueprint, abort, current_app, g, jsonify, make_response, request
from google.auth import jwt
from pydantic import ValidationError
from requests.exceptions import HTTPError
from sqlalchemy.orm import Query
from werkzeug.utils import secure_filename

from cg.store.models import Flowcell

LOG = logging.getLogger(__name__)
BLUEPRINT = Blueprint("api", __name__, url_prefix="/api/v1")


def public(route_function):
    @wraps(route_function)
    def public_endpoint(*args, **kwargs):
        return route_function(*args, **kwargs)

    public_endpoint.is_public = True
    return public_endpoint


@BLUEPRINT.before_request
def before_request():
    """Authorize API routes with JSON Web Tokens."""
    if not request.is_secure:
        return abort(
            make_response(
                jsonify(message="Only https requests accepted"), http.HTTPStatus.FORBIDDEN
            )
        )

    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), http.HTTPStatus.NO_CONTENT)

    endpoint_func = current_app.view_functions[request.endpoint]
    if getattr(endpoint_func, "is_public", None):
        return

    auth_header = request.headers.get("Authorization")
    if not auth_header:
        return abort(
            make_response(
                jsonify(message="no JWT token found on request"), http.HTTPStatus.UNAUTHORIZED
            )
        )

    jwt_token = auth_header.split("Bearer ")[-1]
    try:
        user_data = jwt.decode(
            jwt_token, certs=requests.get("https://www.googleapis.com/oauth2/v1/certs").json()
        )
    except ValueError:
        return abort(
            make_response(
                jsonify(message="outdated login certificate"), http.HTTPStatus.UNAUTHORIZED
            )
        )

    user_obj = db.user(user_data["email"])
    if user_obj is None or not user_obj.order_portal_login:
        message = f"{user_data['email']} doesn't have access"
        LOG.error(message)
        return abort(make_response(jsonify(message=message), http.HTTPStatus.FORBIDDEN))

    g.current_user = user_obj


@BLUEPRINT.route("/submit_order/<order_type>", methods=["POST"])
def submit_order(order_type):
    """Submit an order for samples."""
    api = OrdersAPI(lims=lims, status=db, osticket=osticket)
    error_message: str
    try:
        request_json = request.get_json()
        LOG.info(
            "processing order: %s",
            WriteStream.write_stream_from_content(
                content=request_json, file_format=FileFormat.JSON
            ),
        )
        project: OrderType = OrderType(order_type)
        order_in: OrderIn = OrderIn.parse_obj(request_json, project=project)

        result = api.submit(
            project=project,
            order_in=order_in,
            user_name=g.current_user.name,
            user_mail=g.current_user.email,
        )
    except (  # user misbehaviour
        OrderError,
        OrderFormError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.BAD_REQUEST
        LOG.error(error_message)
    except (  # system misbehaviour
        AttributeError,
        ConnectionError,
        HTTPError,
        IntegrityError,
        KeyError,
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TicketCreationError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(
            project=result["project"], records=[record.to_dict() for record in result["records"]]
        )

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))


@BLUEPRINT.route("/cases")
def cases():
    """Fetch cases."""
    records = db.cases(days=31)
    count = len(records)
    return jsonify(cases=records, total=count)


@BLUEPRINT.route("/families")
def families():
    """Return families."""
    if request.args.get("status") == "analysis":
        records = db.cases_to_mip_analyze()
        count = len(records)
    else:
        customers: Optional[List[Customer]] = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        case_query = db.families(
            enquiry=request.args.get("enquiry"),
            customers=customers,
            action=request.args.get("action"),
        )
        count = case_query.count()
        records = case_query.limit(30)

    cases_data: List[dict] = [case_obj.to_dict(links=True) for case_obj in records]
    return jsonify(families=cases_data, total=count)


@BLUEPRINT.route("/families_in_collaboration")
def families_in_collaboration():
    """Fetch families in collaboration."""
    order_customer: Customer = db.customer(request.args.get("customer"))
    data_analysis: str = request.args.get("data_analysis")
    families_q: Query = db.families(
        enquiry=request.args.get("enquiry"),
        customers=order_customer.collaborators,
        data_analysis=data_analysis,
    )
    count = families_q.count()
    records = families_q.limit(30)
    data = [case_obj.to_dict(links=True) for case_obj in records]
    return jsonify(families=data, total=count)


@BLUEPRINT.route("/families/<family_id>")
def family(family_id):
    """Fetch a family with links."""
    case_obj = db.family(family_id)
    if case_obj is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (case_obj.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)

    data = case_obj.to_dict(links=True, analyses=True)
    return jsonify(**data)


@BLUEPRINT.route("/families_in_collaboration/<family_id>")
def family_in_collaboration(family_id):
    """Fetch a family with links."""
    case_obj = db.family(family_id)
    order_customer = db.customer(request.args.get("customer"))
    if case_obj.customer not in order_customer.collaborators:
        return abort(http.HTTPStatus.FORBIDDEN)
    data = case_obj.to_dict(links=True, analyses=True)
    return jsonify(**data)


@BLUEPRINT.route("/samples")
def samples():
    """Fetch samples."""
    if request.args.get("status") and not g.current_user.is_admin:
        return abort(http.HTTPStatus.FORBIDDEN)
    if request.args.get("status") == "incoming":
        samples_q = db.samples_to_receive()
    elif request.args.get("status") == "labprep":
        samples_q = db.samples_to_prepare()
    elif request.args.get("status") == "sequencing":
        samples_q = db.samples_to_sequence()
    else:
        customer_objs: Optional[Customer] = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        samples_q = db.samples(enquiry=request.args.get("enquiry"), customers=customer_objs)
    limit = int(request.args.get("limit", 50))
    data = [sample_obj.to_dict() for sample_obj in samples_q.limit(limit)]
    return jsonify(samples=data, total=samples_q.count())


@BLUEPRINT.route("/samples_in_collaboration")
def samples_in_collaboration():
    """Fetch samples in a customer group."""
    order_customer = db.customer(request.args.get("customer"))
    samples_q = db.samples(
        enquiry=request.args.get("enquiry"), customers=order_customer.collaborators
    )
    limit = int(request.args.get("limit", 50))
    data = [sample_obj.to_dict() for sample_obj in samples_q.limit(limit)]
    return jsonify(samples=data, total=samples_q.count())


@BLUEPRINT.route("/samples/<sample_id>")
def sample(sample_id):
    """Fetch a single sample."""
    sample_obj = db.sample(sample_id)
    if sample_obj is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (sample_obj.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)
    data = sample_obj.to_dict(links=True, flowcells=True)
    return jsonify(**data)


@BLUEPRINT.route("/samples_in_collaboration/<sample_id>")
def sample_in_collaboration(sample_id):
    """Fetch a single sample."""
    sample_obj = db.sample(sample_id)
    order_customer = db.customer(request.args.get("customer"))
    if sample_obj.customer not in order_customer.collaborators:
        return abort(http.HTTPStatus.FORBIDDEN)
    data = sample_obj.to_dict(links=True, flowcells=True)
    return jsonify(**data)


@BLUEPRINT.route("/pools")
def pools():
    """Fetch pools."""
    customer_objs: Optional[Customer] = (
        None if g.current_user.is_admin else g.current_user.customers
    )
    pools_q = db.pools(customers=customer_objs, enquiry=request.args.get("enquiry"))
    data = [pool_obj.to_dict() for pool_obj in pools_q.limit(30)]
    return jsonify(pools=data, total=pools_q.count())


@BLUEPRINT.route("/pools/<pool_id>")
def pool(pool_id):
    """Fetch a single pool."""
    record = db.pool(pool_id)
    if record is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (record.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**record.to_dict())


@BLUEPRINT.route("/flowcells")
def flowcells() -> Any:
    """Fetch flow cells."""
    flow_cells: List[Flowcell] = db.get_flow_cell_by_enquiry_and_status(
        flow_cell_statuses=[request.args.get("status")],
        flow_cell_id_enquiry=request.args.get("enquiry"),
    )
    parsed_flow_cells: List[dict] = [flow_cell.to_dict() for flow_cell in flow_cells.limit(50)]
    return jsonify(flowcells=parsed_flow_cells, total=flow_cells.count())


@BLUEPRINT.route("/flowcells/<flowcell_id>")
def flowcell(flowcell_id):
    """Fetch a single flowcell."""
    record = db.get_flow_cell(flowcell_id)
    if record is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    return jsonify(**record.to_dict(samples=True))


@BLUEPRINT.route("/analyses")
def analyses():
    """Fetch analyses."""
    if request.args.get("status") == "delivery":
        analyses_q = db.analyses_to_deliver()
    elif request.args.get("status") == "upload":
        analyses_q = db.analyses_to_upload()
    else:
        analyses_q = db.Analysis.query
    data = [analysis_obj.to_dict() for analysis_obj in analyses_q.limit(30)]
    return jsonify(analyses=data, total=analyses_q.count())


@BLUEPRINT.route("/options")
def options():
    """Fetch various options."""
    customers: Optional[Customer] = (
        db.Customer.query.all() if g.current_user.is_admin else g.current_user.customers
    )

    apptag_groups: Dict[str, List[str]] = {"ext": []}
    for application in db.applications(archived=False):
        if not application.versions:
            LOG.debug(f"Skipping application {application} that doesn't have a price")
            continue
        if application.is_external:
            apptag_groups["ext"].append(application.tag)
        if application.prep_category not in apptag_groups:
            apptag_groups[application.prep_category]: List[str] = []
        apptag_groups[application.prep_category].append(application.tag)

    source_groups = {"metagenome": METAGENOME_SOURCES, "analysis": ANALYSIS_SOURCES}

    return jsonify(
        customers=[
            {"text": f"{customer.name} ({customer.internal_id})", "value": customer.internal_id}
            for customer in customers
        ],
        applications=apptag_groups,
        panels=[panel.abbrev for panel in db.panels()],
        organisms=[
            {
                "name": organism.name,
                "reference_genome": organism.reference_genome,
                "internal_id": organism.internal_id,
                "verified": organism.verified,
            }
            for organism in db.organisms()
        ],
        sources=source_groups,
        beds=[bed.name for bed in db.get_active_beds()],
    )


@BLUEPRINT.route("/me")
def me():
    """Fetch information about current user."""
    if not g.current_user.is_admin and not g.current_user.customers:
        LOG.error(
            "%s is not admin and is not connected to any customers, aborting", g.current_user.email
        )
        return abort(http.HTTPStatus.FORBIDDEN)

    return jsonify(user=g.current_user.to_dict())


@BLUEPRINT.route("/applications")
@public
def applications():
    """Fetch application tags."""
    query = db.applications(archived=False)
    data = [record.to_dict() for record in query]
    return jsonify(applications=data)


@BLUEPRINT.route("/applications/<tag>")
@public
def application(tag):
    """Fetch an application tag."""
    record = db.application(tag)
    if record is None:
        return abort(
            make_response(jsonify(message="application not found"), http.HTTPStatus.NOT_FOUND)
        )
    return jsonify(**record.to_dict())


@BLUEPRINT.route("/orderform", methods=["POST"])
def orderform():
    """Parse an orderform/JSON export."""
    input_file = request.files.get("file")
    filename = secure_filename(input_file.filename)

    error_message: str
    try:
        if filename.lower().endswith(".xlsx"):
            temp_dir = Path(tempfile.gettempdir())
            saved_path = str(temp_dir / filename)
            input_file.save(saved_path)
            order_parser = ExcelOrderformParser()
            order_parser.parse_orderform(excel_path=saved_path)
        else:
            json_data = json.load(input_file.stream, strict=False)
            order_parser = JsonOrderformParser()
            order_parser.parse_orderform(order_data=json_data)
        parsed_order: Orderform = order_parser.generate_orderform()
    except (  # user misbehaviour
        AttributeError,
        OrderFormError,
        OverflowError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        LOG.error(error_message)
        http_error_response = http.HTTPStatus.BAD_REQUEST
    except (  # system misbehaviour
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(**parsed_order.dict())

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))



"""Constants specific for encryption"""

from cgmodels.cg.constants import StrEnum

from cg.utils.enums import ListEnum


class CipherAlgorithm(StrEnum):
    TRIPLE_DES: str = "3DES"
    AES: str = "AES"
    AES192: str = "AES192"
    AES256: str = "AES256"
    BLOWFISH: str = "BLOWFISH"
    CAMELLIA128: str = "CAMELLIA128"
    CAMELLIA192: str = "CAMELLIA192"
    CAMELLIA256: str = "CAMELLIA256"
    CAST5: str = "CAST5"
    IDEA: str = "IDEA"
    TWOFISH: str = "TWOFISH"


class EncryptionUserID(StrEnum):
    HASTA_USER_ID: str = "Clinical Genomics"


class GPGParameters(ListEnum):
    ASYMMETRIC_ENCRYPTION: list = [
        "--encrypt",
        "--recipient",
        EncryptionUserID.HASTA_USER_ID,
    ]
    ASYMMETRIC_DECRYPTION: list = [
        "--decrypt",
        "--batch",
        "--cipher-algo",
        CipherAlgorithm.AES256,
        "--passphrase",
        EncryptionUserID.HASTA_USER_ID,
    ]
    SYMMETRIC_ENCRYPTION: list = [
        "--symmetric",
        "--cipher-algo",
        CipherAlgorithm.AES256,
        "--batch",
        "--compress-algo",
        "None",
        "--passphrase-file",
    ]
    SYMMETRIC_DECRYPTION: list = [
        "--decrypt",
        "--cipher-algo",
        CipherAlgorithm.AES256,
        "--batch",
        "--passphrase-file",
    ]
    OUTPUT_PARAMETER: list = [
        "-o",
    ]
